# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains the Simulation class, used for simulation of a Sequence."""

from __future__ import annotations

from typing import Optional, Union, cast, Any
from collections.abc import Mapping
import itertools
from collections import Counter
from copy import deepcopy
from dataclasses import asdict
import warnings

import qutip
import numpy as np
from numpy.typing import ArrayLike

from pulser import Pulse, Sequence
from pulser.simulation.simresults import (
    SimulationResults,
    CoherentResults,
    NoisyResults,
)
from pulser.simulation.simconfig import SimConfig
from pulser._seq_drawer import draw_sequence
from pulser.sequence import _TimeSlot


SUPPORTED_BASES = {"ground-rydberg", "digital"}


class Simulation:
    r"""Simulation of a pulse sequence using QuTiP.

    Args:
        sequence (Sequence): An instance of a Pulser Sequence that we
            want to simulate.
        sampling_rate (float): The fraction of samples that we wish to
            extract from the pulse sequence to simulate. Has to be a
            value between 0.05 and 1.0.
        config (SimConfig): Configuration to be used for this simulation.
        evaluation_times (Union[str, ArrayLike, float]): Choose between:

            - "Full": The times are set to be the ones used to define the
              Hamiltonian to the solver.

            - "Minimal": The times are set to only include initial and final
              times.

            - An ArrayLike object of times in µs if you wish to only include
              those specific times.

            - A float to act as a sampling rate for the resulting state.
    """

    def __init__(
        self,
        sequence: Sequence,
        sampling_rate: float = 1.0,
        config: Optional[SimConfig] = None,
        evaluation_times: Union[float, str, ArrayLike] = "Full",
    ) -> None:
        """Instantiates a Simulation object."""
        if not isinstance(sequence, Sequence):
            raise TypeError(
                "The provided sequence has to be a valid "
                "pulser.Sequence instance."
            )
        if not sequence._schedule:
            raise ValueError("The provided sequence has no declared channels.")
        if all(sequence._schedule[x][-1].tf == 0 for x in sequence._channels):
            raise ValueError(
                "No instructions given for the channels in the " "sequence."
            )
        not_supported = (
            set(ch.basis for ch in sequence._channels.values())
            - SUPPORTED_BASES
        )
        if not_supported:
            raise NotImplementedError(
                "Sequence with unsupported bases: " + "".join(not_supported)
            )
        self._seq = sequence
        self._qdict = self._seq.qubit_info
        self._size = len(self._qdict)
        self._tot_duration = self._seq.get_duration()
        if not (0 < sampling_rate <= 1.0):
            raise ValueError(
                "The sampling rate (`sampling_rate` = "
                f"{sampling_rate}) must be greater than 0 and "
                "less than or equal to 1."
            )
        if int(self._tot_duration * sampling_rate) < 4:
            raise ValueError(
                "`sampling_rate` is too small, less than 4 data " "points."
            )
        self._sampling_rate = sampling_rate
        self._qid_index = {qid: i for i, qid in enumerate(self._qdict)}
        self._collapse_ops: list[qutip.Qobj] = []
        self._times = self._adapt_to_sampling_rate(
            np.arange(self._tot_duration, dtype=np.double) / 1000
        )
        self.evaluation_times = evaluation_times
        self._bad_atoms: dict[Union[str, int], bool] = {}
        self._doppler_detune: dict[Union[str, int], float] = {}
        # Sets the config as well as builds the hamiltonian
        self.set_config(config) if config else self.set_config(SimConfig())
        if hasattr(self._seq, "_measurement"):
            self._meas_basis = cast(str, self._seq._measurement)
        else:
            if self.basis_name in {"digital", "all"}:
                self._meas_basis = "digital"
            else:
                self._meas_basis = "ground-rydberg"
        self.initial_state = "all-ground"

    @property
    def config(self) -> SimConfig:
        """The current configuration, as a SimConfig instance."""
        return self._config

    def set_config(self, cfg: SimConfig) -> None:
        """Sets current config to cfg and updates simulation parameters.

        Args:
            cfg (SimConfig): New configuration.
        """
        if not isinstance(cfg, SimConfig):
            raise ValueError(f"Object {cfg} is not a valid `SimConfig`.")
        self._config = cfg
        if not ("SPAM" in self.config.noise and self.config.eta > 0):
            self._bad_atoms = {qid: False for qid in self._qid_index}
        if "doppler" not in self.config.noise:
            self._doppler_detune = {qid: 0.0 for qid in self._qid_index}
        # Noise, samples and Hamiltonian update routine
        self._construct_hamiltonian()
        if "dephasing" in self.config.noise:
            if self.basis_name == "digital" or self.basis_name == "all":
                raise NotImplementedError(
                    "Cannot include dephasing noise in"
                    + " digital- or all-basis."
                )
            # Probability of phase (Z) flip:
            # First order in prob
            prob = self.config.dephasing_prob / 2
            n = self._size
            if prob > 0.1 and n > 1:
                warnings.warn(
                    "The dephasing model is a first-order approximation in the"
                    + f" dephasing probability. p = {2*prob} is too large for "
                    "realistic results.",
                    stacklevel=2,
                )
            k = np.sqrt(prob * (1 - prob) ** (n - 1))
            self._collapse_ops = [
                np.sqrt((1 - prob) ** n)
                * qutip.tensor([self.op_matrix["I"] for _ in range(n)])
            ]
            self._collapse_ops += [
                k
                * (
                    self._build_operator("sigma_rr", qid)
                    - self._build_operator("sigma_gg", qid)
                )
                for qid in self._qid_index
            ]

    def add_config(self, config: SimConfig) -> None:
        """Updates the current configuration with parameters of another one.

        Mostly useful when dealing with multiple noise types in different
        configurations and wanting to merge these configurations together.
        Adds simulation parameters to noises that weren't available in the
        former SimConfig. Noises specified in both SimConfigs will keep
        former noise parameters.

        Args:
            config (SimConfig): SimConfig to retrieve parameters from.
        """
        if not isinstance(config, SimConfig):
            raise ValueError(f"Object {config} is not a valid `SimConfig`")

        old_noise_set = set(self.config.noise)
        new_noise_set = old_noise_set.union(config.noise)
        diff_noise_set = new_noise_set - old_noise_set
        param_dict: dict[str, Any] = asdict(self._config)
        del param_dict["spam_dict"]
        del param_dict["doppler_sigma"]
        param_dict["noise"] = tuple(new_noise_set)
        if "SPAM" in diff_noise_set:
            param_dict["eta"] = config.eta
            param_dict["epsilon"] = config.epsilon
            param_dict["epsilon_prime"] = config.epsilon_prime
        if "doppler" in diff_noise_set:
            param_dict["temperature"] = config.temperature
        if "amplitude" in diff_noise_set:
            param_dict["laser_waist"] = config.laser_waist
        if "dephasing" in diff_noise_set:
            param_dict["dephasing_prob"] = config.dephasing_prob
        param_dict["temperature"] *= 1.0e6
        self.set_config(SimConfig(**param_dict))

    def show_config(self, solver_options: bool = False) -> None:
        """Shows current configuration."""
        print(self._config.__str__(solver_options))

    def reset_config(self) -> None:
        """Resets configuration to default."""
        self.set_config(SimConfig())

    @property
    def initial_state(self) -> qutip.Qobj:
        """The initial state of the simulation.

        Args:
            state (Union[str, ArrayLike, qutip.Qobj]): The initial state.
                Choose between:

                - "all-ground" for all atoms in ground state
                - An ArrayLike with a shape compatible with the system
                - A Qobj object
        """
        return self._initial_state

    @initial_state.setter
    def initial_state(self, state: Union[str, np.ndarray, qutip.Qobj]) -> None:
        """Sets the initial state of the simulation."""
        self._initial_state: qutip.Qobj
        if state == "all-ground":
            self._initial_state = qutip.tensor(
                [self.basis["g"] for _ in range(self._size)]
            )
        else:
            state = cast(Union[np.ndarray, qutip.Qobj], state)
            shape = state.shape[0]
            legal_shape = self.dim ** self._size
            legal_dims = [[self.dim] * self._size, [1] * self._size]
            if shape != legal_shape:
                raise ValueError(
                    "Incompatible shape of initial state."
                    + f"Expected {legal_shape}, got {shape}."
                )
            self._initial_state = qutip.Qobj(state, dims=legal_dims)

    @property
    def evaluation_times(self) -> Union[str, float, ArrayLike]:
        """The times at which the results of this simulation are returned.

        Args:
            value (Union[str, ArrayLike, float]): Choose between:

                - "Full": The times are set to be the ones used to define the
                  Hamiltonian to the solver.

                - "Minimal": The times are set to only include initial and
                  final times.

                - An ArrayLike object of times in µs if you wish to only
                  include those specific times.

                - A float to act as a sampling rate for the resulting state.
        """
        return self._evaluation_times

    @evaluation_times.setter
    def evaluation_times(self, value: Union[str, ArrayLike, float]) -> None:
        """Sets times at which the results of this simulation are returned."""
        if isinstance(value, str):
            if value == "Full":
                self._eval_times_array = self._times
            elif value == "Minimal":
                self._eval_times_array = np.array(
                    [self._times[0], self._times[-1]]
                )
            else:
                raise ValueError(
                    "Wrong evaluation time label. It should "
                    "be `Full`, `Minimal`, an array of times or"
                    + " a float between 0 and 1."
                )
        elif isinstance(value, float):
            if value > 1 or value <= 0:
                raise ValueError(
                    "evaluation_times float must be between 0 " "and 1."
                )
            indices = np.linspace(
                0,
                len(self._times) - 1,
                int(value * len(self._times)),
                dtype=int,
            )
            self._eval_times_array = self._times[indices]
        elif isinstance(value, (list, tuple, np.ndarray)):
            t_max = np.max(value)
            t_min = np.min(value)
            if t_max > self._times[-1]:
                raise ValueError(
                    "Provided evaluation-time list extends "
                    "further than sequence duration."
                )
            if t_min < 0:
                raise ValueError(
                    "Provided evaluation-time list contains "
                    "negative values."
                )
            # Ensure the list of times is sorted
            eval_times = np.array(np.sort(value))
            if t_min > 0:
                eval_times = np.insert(eval_times, 0, 0.0)
            if t_max < self._times[-1]:
                eval_times = np.append(eval_times, self._times[-1])
            self._eval_times_array = eval_times
            # always include initial and final times
        else:
            raise ValueError(
                "Wrong evaluation time label. It should "
                "be `Full`, `Minimal`, an array of times or a "
                + "float between 0 and 1."
            )
        self._evaluation_times: Union[str, ArrayLike, float] = value

    def draw(
        self,
        draw_phase_area: bool = False,
        draw_interp_pts: bool = False,
        draw_phase_shifts: bool = False,
    ) -> None:
        """Draws the input sequence and the one used by the solver.

        Keyword Args:
            draw_phase_area (bool): Whether phase and area values need
                to be shown as text on the plot, defaults to False.
            draw_interp_pts (bool): When the sequence has pulses with waveforms
                of type InterpolatedWaveform, draws the points of interpolation
                on top of the respective waveforms (defaults to False).
            draw_phase_shifts (bool): Whether phase shift and reference
                information should be added to the plot, defaults to False.

        See Also:
            Sequence.draw(): Draws the sequence in its current state.
        """
        draw_sequence(
            self._seq,
            self._sampling_rate,
            draw_phase_area=draw_phase_area,
            draw_interp_pts=draw_interp_pts,
            draw_phase_shifts=draw_phase_shifts,
        )

    def _extract_samples(self) -> None:
        """Populates samples dictionary with every pulse in the sequence."""
        self.samples: dict[str, dict[str, dict]] = {
            addr: {basis: {} for basis in ["ground-rydberg", "digital"]}
            for addr in ["Global", "Local"]
        }
        if not hasattr(self, "operators"):
            self.operators = deepcopy(self.samples)

        def prepare_dict() -> dict[str, np.ndarray]:
            # Duration includes retargeting, delays, etc.
            return {
                "amp": np.zeros(self._tot_duration),
                "det": np.zeros(self._tot_duration),
                "phase": np.zeros(self._tot_duration),
            }

        def write_samples(
            slot: _TimeSlot,
            samples_dict: Mapping[str, np.ndarray],
            is_global_pulse: bool,
            *qid: Union[int, str],
        ) -> None:
            """Builds hamiltonian coefficients.

            Taking into account, if necessary, noise effects, which are local
            and depend on the qubit's id qid.
            """
            _pulse = cast(Pulse, slot.type)
            noise_det = 0.0
            noise_amp = 1.0
            if "doppler" in self.config.noise:
                noise_det += self._doppler_detune[qid[0]]
            # Gaussian beam loss in amplitude for global pulses only
            # Noise is drawn at random for each pulse
            if "amplitude" in self.config.noise and is_global_pulse:
                position = self._qdict[qid[0]]
                r = np.linalg.norm(position)
                w0 = self.config.laser_waist
                noise_amp = np.random.normal(1.0, 1.0e-3) * np.exp(
                    -((r / w0) ** 2)
                )
            samples_dict["amp"][slot.ti : slot.tf] = (
                _pulse.amplitude.samples * noise_amp
            )
            samples_dict["det"][slot.ti : slot.tf] = (
                _pulse.detuning.samples + noise_det
            )
            samples_dict["phase"][slot.ti : slot.tf] = _pulse.phase

        for channel in self._seq.declared_channels:
            addr = self._seq.declared_channels[channel].addressing
            basis = self._seq.declared_channels[channel].basis

            # Case of coherent global simulations
            if addr == "Global" and (
                set(self.config.noise).issubset({"dephasing"})
            ):
                samples_dict = self.samples["Global"][basis]
                if not samples_dict:
                    samples_dict = prepare_dict()
                for slot in self._seq._schedule[channel]:
                    if isinstance(slot.type, Pulse):
                        write_samples(slot, samples_dict, True)
                self.samples["Global"][basis] = samples_dict

            # Any noise : global becomes local for each qubit in the reg
            # Since coefficients are modified locally by all noises
            else:
                is_global = addr == "Global"
                samples_dict = self.samples["Local"][basis]
                for slot in self._seq._schedule[channel]:
                    if isinstance(slot.type, Pulse):
                        for qubit in slot.targets:
                            if qubit not in samples_dict:
                                samples_dict[qubit] = prepare_dict()
                            # We don't write samples for badly prep qubits
                            if not self._bad_atoms[qubit]:
                                write_samples(
                                    slot, samples_dict[qubit], is_global, qubit
                                )
                self.samples["Local"][basis] = samples_dict

    def _build_operator(
        self, op_id: str, *qubit_ids: Union[str, int], global_op: bool = False
    ) -> qutip.Qobj:
        """Create qutip.Qobj with nontrivial action at *qubit_ids."""
        if global_op:
            return sum(
                self._build_operator(op_id, q_id) for q_id in self._qdict
            )
        if len(set(qubit_ids)) < len(qubit_ids):
            raise ValueError("Duplicate atom ids in argument list.")
        # List of identity operators, except for op_id where requested:
        op_list = [
            self.op_matrix[op_id]
            if j in map(self._qid_index.get, qubit_ids)
            else self.op_matrix["I"]
            for j in range(self._size)
        ]
        return qutip.tensor(op_list)

    def _adapt_to_sampling_rate(self, full_array: np.ndarray) -> np.ndarray:
        """Adapt list to correspond to sampling rate."""
        indices = np.linspace(
            0,
            self._tot_duration - 1,
            int(self._sampling_rate * self._tot_duration),
            dtype=int,
        )
        return cast(np.ndarray, full_array[indices])

    def _update_noise(self) -> None:
        """Updates noise random parameters.

        Used at the start of each run. If SPAM isn't in chosen noises, all
        atoms are set to be correctly prepared.
        """
        if "SPAM" in self.config.noise and self.config.eta > 0:
            dist = (
                np.random.uniform(size=len(self._qid_index))
                < self.config.spam_dict["eta"]
            )
            self._bad_atoms = dict(zip(self._qid_index, dist))
        if "doppler" in self.config.noise:
            detune = np.random.normal(
                0, self.config.doppler_sigma, size=len(self._qid_index)
            )
            self._doppler_detune = dict(zip(self._qid_index, detune))

    def _build_basis_and_op_matrices(self) -> None:
        """Determine dimension, basis and projector operators."""
        # No samples => Empty dict entry => False
        if (
            not self.samples["Global"]["digital"]
            and not self.samples["Local"]["digital"]
        ):
            self.basis_name = "ground-rydberg"
            self.dim = 2
            basis = ["r", "g"]
            projectors = ["gr", "rr", "gg"]
        elif (
            not self.samples["Global"]["ground-rydberg"]
            and not self.samples["Local"]["ground-rydberg"]
        ):
            self.basis_name = "digital"
            self.dim = 2
            basis = ["g", "h"]
            projectors = ["hg", "hh", "gg"]
        else:
            self.basis_name = "all"  # All three states
            self.dim = 3
            basis = ["r", "g", "h"]
            projectors = ["gr", "hg", "rr", "gg", "hh"]

        self.basis = {b: qutip.basis(self.dim, i) for i, b in enumerate(basis)}
        self.op_matrix = {"I": qutip.qeye(self.dim)}

        for proj in projectors:
            self.op_matrix["sigma_" + proj] = (
                self.basis[proj[0]] * self.basis[proj[1]].dag()
            )

    def _construct_hamiltonian(self) -> None:
        """Constructs the hamiltonian from the Sequence.

        Also builds qutip.Qobjs related to the Sequence if not built already,
        and refreshes potential noise parameters by drawing new at random.
        """
        self._update_noise()
        self._extract_samples()
        if not hasattr(self, "basis_name"):
            self._build_basis_and_op_matrices()

        def make_vdw_term() -> qutip.Qobj:
            """Construct the Van der Waals interaction Term.

            For each pair of qubits, calculate the distance between them, then
            assign the local operator "sigma_rr" at each pair. The units are
            given so that the coefficient includes a 1/hbar factor.
            """
            vdw = 0 * self._build_operator("I")
            # Get every pair without duplicates
            for q1, q2 in itertools.combinations(self._qdict.keys(), r=2):
                # no VdW interaction with other qubits for a badly prep. qubit
                if not (self._bad_atoms[q1] or self._bad_atoms[q2]):
                    dist = np.linalg.norm(self._qdict[q1] - self._qdict[q2])
                    U = 0.5 * self._seq._device.interaction_coeff / dist ** 6
                    vdw += U * self._build_operator("sigma_rr", q1, q2)
            return vdw

        def build_coeffs_ops(basis: str, addr: str) -> list[list]:
            """Build coefficients and operators for the hamiltonian QobjEvo."""
            samples = self.samples[addr][basis]
            operators = self.operators[addr][basis]
            # Choose operator names according to addressing:
            if basis == "ground-rydberg":
                op_ids = ["sigma_gr", "sigma_rr"]
            elif basis == "digital":
                op_ids = ["sigma_hg", "sigma_gg"]

            terms = []
            if addr == "Global":
                coeffs = [
                    0.5 * samples["amp"] * np.exp(-1j * samples["phase"]),
                    -0.5 * samples["det"],
                ]
                for op_id, coeff in zip(op_ids, coeffs):
                    if np.any(coeff != 0):
                        # Build once global operators as they are needed
                        if op_id not in operators:
                            operators[op_id] = self._build_operator(
                                op_id, global_op=True
                            )
                        terms.append(
                            [
                                operators[op_id],
                                self._adapt_to_sampling_rate(coeff),
                            ]
                        )
            elif addr == "Local":
                for q_id, samples_q in samples.items():
                    if q_id not in operators:
                        operators[q_id] = {}
                    coeffs = [
                        0.5
                        * samples_q["amp"]
                        * np.exp(-1j * samples_q["phase"]),
                        -0.5 * samples_q["det"],
                    ]
                    for coeff, op_id in zip(coeffs, op_ids):
                        if np.any(coeff != 0):
                            if op_id not in operators[q_id]:
                                operators[q_id][op_id] = self._build_operator(
                                    op_id, q_id
                                )
                            terms.append(
                                [
                                    operators[q_id][op_id],
                                    self._adapt_to_sampling_rate(coeff),
                                ]
                            )
            self.operators[addr][basis] = operators
            return terms

        # Time independent term:
        if self.basis_name == "digital" or self._size == 1:
            qobj_list = [0 * self._build_operator("I")]
        else:
            # Van der Waals Interaction Terms
            qobj_list = [make_vdw_term()]

        # Time dependent terms:
        for addr in self.samples:
            for basis in self.samples[addr]:
                if self.samples[addr][basis]:
                    qobj_list += cast(list, build_coeffs_ops(basis, addr))

        ham = qutip.QobjEvo(qobj_list, tlist=self._times)
        ham = ham + ham.dag()
        ham.compress()
        self._hamiltonian = ham

    def get_hamiltonian(self, time: float) -> qutip.Qobj:
        """Get the Hamiltonian created from the sequence at a fixed time.

        Args:
            time (float): The specific time at which we want to extract the
                Hamiltonian (in ns).

        Returns:
            qutip.Qobj: A new Qobj for the Hamiltonian with coefficients
            extracted from the effective sequence (determined by
            `self.sampling_rate`) at the specified time.
        """
        if time > 1000 * self._times[-1]:
            raise ValueError(
                f"Provided time (`time` = {time}) must be "
                "less than or equal to the sequence duration "
                f"({1000 * self._times[-1]})."
            )
        if time < 0:
            raise ValueError(
                f"Provided time (`time` = {time}) must be "
                "greater than or equal to 0."
            )
        return self._hamiltonian(time / 1000)  # Creates new Qutip.Qobj

    # Run Simulation Evolution using Qutip
    def run(
        self,
        progress_bar: Optional[bool] = None,
        **options: qutip.solver.Options,
    ) -> SimulationResults:
        """Simulates the sequence using QuTiP's solvers.

        Will return NoisyResults if the noise in the SimConfig requires it.
        Otherwise will return CoherentResults.

        Keyword Args:
            progress_bar (bool): If True, the progress bar of QuTiP's solver
                will be shown.
            options (qutip.solver.Options): If specified, will override
                SimConfig solver_options. If no `max_step` value is provided,
                an automatic one is calculated from the `Sequence`'s schedule
                (half of the shortest duration among pulses and delays).
        """
        if "max_step" in options.keys():
            solv_ops = qutip.Options(**options)
        else:
            auto_max_step = 0.5 * (self._seq._min_pulse_duration() / 1000)
            solv_ops = qutip.Options(max_step=auto_max_step, **options)

        meas_errors: Optional[Mapping[str, float]] = None
        if "SPAM" in self.config.noise:
            meas_errors = {
                k: self.config.spam_dict[k]
                for k in ("epsilon", "epsilon_prime")
            }

        def _run_solver() -> CoherentResults:
            """Returns CoherentResults: Object containing evolution results."""
            if "dephasing" in self.config.noise:
                # temporary workaround due to a qutip bug when using mesolve
                liouvillian = qutip.liouvillian(
                    self._hamiltonian, self._collapse_ops
                )
                result = qutip.mesolve(
                    liouvillian,
                    self.initial_state,
                    self._eval_times_array,
                    progress_bar=progress_bar,
                    options=solv_ops,
                )
            else:
                result = qutip.sesolve(
                    self._hamiltonian,
                    self.initial_state,
                    self._eval_times_array,
                    progress_bar=progress_bar,
                    options=solv_ops,
                )
            return CoherentResults(
                result.states,
                self._size,
                self.basis_name,
                self._eval_times_array,
                self._meas_basis,
                meas_errors,
            )

        # Check if noises ask for averaging over multiple runs:
        if set(self.config.noise).issubset({"dephasing", "SPAM"}):
            # If there is "SPAM", the preparation errors must be zero
            if "SPAM" not in self.config.noise or self.config.eta == 0:
                return _run_solver()

        # Did not obey the conditions above, will return NoisyResults
        time_indices = range(len(self._eval_times_array))
        total_count = np.array([Counter() for _ in time_indices])
        # We run the system multiple times
        for _ in range(self.config.runs):
            # At each run, new random noise: new Hamiltonian
            self._construct_hamiltonian()
            # Get CoherentResults instance from sequence with added noise:
            cleanres_noisyseq = _run_solver()
            # Extract statistics at eval time:
            total_count += np.array(
                [
                    cleanres_noisyseq.sample_state(
                        t, n_samples=self.config.samples_per_run
                    )
                    for t in self._eval_times_array
                ]
            )
        n_measures = self.config.runs * self.config.samples_per_run
        total_run_prob = [
            Counter({k: v / n_measures for k, v in total_count[t].items()})
            for t in time_indices
        ]
        return NoisyResults(
            total_run_prob,
            self._size,
            self.basis_name,
            self._eval_times_array,
            n_measures,
        )



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pulser.devices._device_datacls import Device
from pulser.channels import Rydberg, Raman, Microwave


MockDevice = Device(
    name="MockDevice",
    dimensions=2,
    max_atom_num=2000,
    max_radial_distance=1000,
    min_atom_distance=1,
    _channels=(
        (
            "rydberg_global",
            Rydberg.Global(1000, 200, clock_period=1, min_duration=1),
        ),
        (
            "rydberg_local",
            Rydberg.Local(
                1000, 200, 0, max_targets=2000, clock_period=1, min_duration=1
            ),
        ),
        (
            "raman_global",
            Raman.Global(1000, 200, clock_period=1, min_duration=1),
        ),
        (
            "raman_local",
            Raman.Local(
                1000, 200, 0, max_targets=2000, clock_period=1, min_duration=1
            ),
        ),
        (
            "mw_global",
            Microwave.Global(1000, 200, clock_period=1, min_duration=1),
        ),
    ),
)



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains the SimConfig class that sets the configuration of a simulation."""

from __future__ import annotations

from sys import version_info
from dataclasses import dataclass, field
from typing import Union

import numpy as np
import qutip

if version_info[:2] >= (3, 8):  # pragma: no cover
    from typing import Literal, get_args
else:  # pragma: no cover
    try:
        from typing_extensions import Literal, get_args  # type: ignore
    except ImportError:
        raise ImportError(
            "Using pulser with Python version 3.7 requires the"
            " `typing_extensions` module. Install it by running"
            " `pip install typing-extensions`."
        )

NOISE_TYPES = Literal["doppler", "amplitude", "SPAM", "dephasing"]
MASS = 1.45e-25  # kg
KB = 1.38e-23  # J/K
KEFF = 8.7  # µm^-1


@dataclass(frozen=True)
class SimConfig:
    """Specifies a simulation's configuration.

    Note:
        Being a frozen dataclass, the configuration chosen upon instantiation
        cannot be changed later on.

    Args:
        noise (Union[str, tuple[str]]): Types of noises to be used in the
            simulation. You may specify just one, or a tuple of the allowed
            noise types:

            - "dephasing": Random phase (Z) flip
            - "doppler": Local atom detuning due to finite speed of the
              atoms and Doppler effect with respect to laser frequency
            - "amplitude": Gaussian damping due to finite laser waist
            - "SPAM": SPAM errors. Defined by **eta**, **epsilon** and
              **epsilon_prime**.

        eta (float): Probability of each atom to be badly prepared.
        epsilon (float): Probability of false positives.
        epsilon_prime(float): Probability of false negatives.
        runs (int): Number of runs needed : each run draws a new random
            noise.
        samples_per_run (int): Number of samples per noisy run.
            Useful for cutting down on computing time, but unrealistic.
        temperature (float): Temperature, set in µK, of the Rydberg array.
            Also sets the standard deviation of the speed of the atoms.
        laser_waist (float): Waist of the gaussian laser, set in µm,
            in global pulses.
        solver_options (qutip.Options): Options for the qutip solver.
    """

    noise: Union[NOISE_TYPES, tuple[NOISE_TYPES, ...]] = ()
    runs: int = 15
    samples_per_run: int = 5
    temperature: float = 50.0
    laser_waist: float = 175.0
    eta: float = 0.005
    epsilon: float = 0.01
    epsilon_prime: float = 0.05
    dephasing_prob: float = 0.05
    solver_options: qutip.Options = None
    spam_dict: dict[str, float] = field(
        init=False, default_factory=dict, repr=False
    )
    doppler_sigma: float = field(
        init=False, default=KEFF * np.sqrt(KB * 50.0e-6 / MASS)
    )

    def __post_init__(self) -> None:
        self._process_temperature()
        self.__dict__["spam_dict"] = {
            "eta": self.eta,
            "epsilon": self.epsilon,
            "epsilon_prime": self.epsilon_prime,
        }
        self._check_noise_types()
        self._check_spam_dict()
        self._calc_sigma_doppler()

    def __str__(self, solver_options: bool = False) -> str:
        lines = [
            "Options:",
            "----------",
            f"Number of runs:        {self.runs}",
            f"Samples per run:       {self.samples_per_run}",
        ]
        if self.noise:
            lines.append("Noise types:           " + ", ".join(self.noise))
        if "SPAM" in self.noise:
            lines.append(f"SPAM dictionary:       {self.spam_dict}")
        if "doppler" in self.noise:
            lines.append(f"Temperature:           {self.temperature*1.e6}µK")
        if "amplitude" in self.noise:
            lines.append(f"Laser waist:           {self.laser_waist}μm")
        if "dephasing" in self.noise:
            lines.append(f"Dephasing probability: {self.dephasing_prob}")
        if solver_options:
            lines.append(
                "Solver Options: \n" + f"{str(self.solver_options)[10:-1]}"
            )
        return "\n".join(lines).rstrip()

    def _check_spam_dict(self) -> None:
        for param, value in self.spam_dict.items():
            if value > 1 or value < 0:
                raise ValueError(
                    f"SPAM parameter {param} = {value} must be"
                    + " greater than 0 and less than 1."
                )

    def _process_temperature(self) -> None:
        # checks value of temperature field and converts it to K from muK
        if self.temperature <= 0:
            raise ValueError(
                "Temperature field"
                + f" (`temperature` = {self.temperature}) must be"
                + " greater than 0."
            )
        self.__dict__["temperature"] *= 1.0e-6

    def _check_noise_types(self) -> None:
        # only one noise was given as argument : convert it to a tuple
        if isinstance(self.noise, str):
            self.__dict__["noise"] = (self.noise,)
        for noise_type in self.noise:
            if noise_type not in get_args(NOISE_TYPES):
                raise ValueError(
                    f"{noise_type} is not a valid noise type. "
                    + "Valid noise types: "
                    + ", ".join(get_args(NOISE_TYPES))
                )

    def _calc_sigma_doppler(self) -> None:
        # sigma = keff Deltav, keff = 8.7mum^-1, Deltav = sqrt(kB T / m)
        self.__dict__["doppler_sigma"] = KEFF * np.sqrt(
            KB * self.temperature / MASS
        )



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains the custom Encoder and Decoder for JSON serialization."""

from __future__ import annotations

import importlib
import inspect
from json import JSONEncoder, JSONDecoder
from typing import Any, cast

import numpy as np

from pulser.json.utils import obj_to_dict
from pulser.parametrized import Variable


class PulserEncoder(JSONEncoder):
    """The custom encoder for Pulser objects."""

    def default(self, o: Any) -> dict[str, Any]:
        """Handles JSON encoding of objects not supported by default."""
        if hasattr(o, "_to_dict"):
            return cast(dict, o._to_dict())
        elif type(o) == type:
            return obj_to_dict(o, _build=False, _name=o.__name__)
        elif isinstance(o, np.ndarray):
            return obj_to_dict(o, o.tolist(), _name="array")
        elif isinstance(o, set):
            return obj_to_dict(o, list(o))
        else:
            return cast(dict, JSONEncoder.default(self, o))


class PulserDecoder(JSONDecoder):
    """The custom decoder for Pulser objects."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initializes the decoder."""
        # TODO: Check version compatibility (stored at the Sequence level)
        self.vars: dict[str, Variable] = {}
        super().__init__(object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj: dict[str, Any]) -> Any:
        """Enforces custom deserializations when decoding."""
        try:
            build = obj["_build"]
            obj_name = obj["__name__"]
            module_str = obj["__module__"]
        except KeyError:
            return obj

        if (
            obj_name == "Variable"
            and module_str == "pulser.parametrized.variable"
        ):
            var_name = obj["name"]
            try:
                var = self.vars[var_name]
                assert var.name == var_name, (
                    f"Variable {var.name} already "
                    + f"declared under {var_name}."
                )
                assert var.dtype == obj["dtype"], (
                    "Mismatching variable types for variables under the name "
                    + f"'{var_name}'."
                )
                assert var.size == obj["size"], (
                    "Mismatching sizes for variables under the name "
                    + f"'{var_name}'."
                )
            except KeyError:
                var = Variable(var_name, obj["dtype"], obj["size"])
                self.vars[var_name] = var

            return var

        module = importlib.import_module(module_str)
        if "__submodule__" in obj:
            submodule = getattr(module, obj["__submodule__"])
            cls = getattr(submodule, obj_name)
            if inspect.ismethod(cls):
                cls = cls.__func__  # Use the unbound function by default
        else:
            cls = getattr(module, obj_name)

        if not build:
            return cls

        if obj_name == "Sequence":
            seq = cls(*obj["__args__"], **obj["__kwargs__"])
            for name, args, kwargs in obj["calls"]:
                getattr(seq, name)(*args, **kwargs)
            seq._building = obj["vars"] == {}
            for name, var in obj["vars"].items():
                assert name not in seq._variables, (
                    "Multiples variables with" + f" the name '{name}'."
                )
                seq._variables[name] = var
            for name, args, kwargs in obj["to_build_calls"]:
                getattr(seq, name)(*args, **kwargs)
            return seq
        else:
            return cls(*obj["__args__"], **obj["__kwargs__"])



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains the ParamObj and auxiliary classes for object parametrization."""

from __future__ import annotations

from collections.abc import Callable
from functools import partialmethod
from itertools import chain
import inspect
import operator
import warnings
from typing import Any, Union, TYPE_CHECKING

from pulser.json.utils import obj_to_dict
from pulser.parametrized import Parametrized

if TYPE_CHECKING:
    from pulser.parametrized import Variable  # pragma: no cover

# Available operations on parametrized objects with OpSupport
reversible_ops = [
    "__add__",
    "__sub__",
    "__mul__",
    "__truediv__",
    "__floordiv__",
    "__pow__",
    "__mod__",
]


class OpSupport:
    """Methods for supporting operators on parametrized objects."""

    def _do_op(self, op_name: str, other: Union[int, float]) -> ParamObj:
        return ParamObj(getattr(operator, op_name), self, other)

    def _do_rop(self, op_name: str, other: Union[int, float]) -> ParamObj:
        return ParamObj(getattr(operator, op_name), other, self)

    def __neg__(self) -> ParamObj:
        return ParamObj(operator.neg, self)

    def __abs__(self) -> ParamObj:
        return ParamObj(operator.abs, self)


# Inject operator magic methods into OpSupport
for method in reversible_ops:
    rmethod = "__r" + method[2:]
    setattr(OpSupport, method, partialmethod(OpSupport._do_op, method))
    setattr(OpSupport, rmethod, partialmethod(OpSupport._do_rop, method))


class ParamObj(Parametrized, OpSupport):
    """Holds a call to a given class.

    When called, a ParamObj instance returns `cls(*args, **kwargs)`.

    Args:
        cls (callable): The object to call. Usually it's a class that's
            instantiated when called.
        args: The args for calling `cls`.
        kwargs: The kwargs for calling `cls`.
    """

    def __init__(self, cls: Callable, *args: Any, **kwargs: Any) -> None:
        """Initializes a new ParamObj."""
        self.cls = cls
        self._variables: dict[str, Variable] = {}
        if isinstance(self.cls, Parametrized):
            self._variables.update(self.cls.variables)
        for x in chain(args, kwargs.values()):
            if isinstance(x, Parametrized):
                self._variables.update(x.variables)
        self.args = args
        self.kwargs = kwargs
        self._instance = None
        self._vars_state: dict[str, int] = {}

    @property
    def variables(self) -> dict[str, Variable]:
        """Returns all involved variables."""
        return self._variables

    def build(self) -> Any:
        """Builds the object with its variables last assigned values."""
        vars_state = {key: var._count for key, var in self._variables.items()}
        if vars_state != self._vars_state:
            self._vars_state = vars_state
            # Builds all Parametrized arguments before feeding them to cls
            args_ = [
                arg.build() if isinstance(arg, Parametrized) else arg
                for arg in self.args
            ]
            kwargs_ = {
                key: val.build() if isinstance(val, Parametrized) else val
                for key, val in self.kwargs.items()
            }
            if isinstance(self.cls, ParamObj):
                obj = self.cls.build()
            else:
                obj = self.cls
            self._instance = obj(*args_, **kwargs_)
        return self._instance

    def _to_dict(self) -> dict[str, Any]:
        def class_to_dict(cls: Callable) -> dict[str, Any]:
            return obj_to_dict(
                self, _build=False, _name=cls.__name__, _module=cls.__module__
            )

        args = list(self.args)
        if isinstance(self.cls, Parametrized):
            cls_dict = self.cls._to_dict()
        elif hasattr(args[0], self.cls.__name__) and inspect.isfunction(
            self.cls
        ):
            # Check for parametrized methods
            if inspect.isclass(self.args[0]):
                # classmethod
                cls_dict = obj_to_dict(
                    self,
                    _build=False,
                    _name=self.cls.__name__,
                    _module=self.args[0].__module__,
                    _submodule=self.args[0].__name__,
                )
                args[0] = class_to_dict(self.args[0])
            else:
                raise NotImplementedError(
                    "Instance or static method "
                    "serialization is not supported."
                )
        else:
            cls_dict = class_to_dict(self.cls)

        return obj_to_dict(self, cls_dict, *args, **self.kwargs)

    def __call__(self, *args: Any, **kwargs: Any) -> ParamObj:
        """Returns a new ParamObj storing a call to the current ParamObj."""
        obj = ParamObj(self, *args, **kwargs)
        warnings.warn(
            "Calls to methods of parametrized objects are only "
            "executed if they serve as arguments of other "
            "parametrized objects that are themselves built. If this"
            f" is not the case, the call to {obj} will not be "
            "executed upon sequence building.",
            stacklevel=2,
        )
        return obj

    def __getattr__(self, name: str) -> ParamObj:
        if hasattr(self.cls, name):
            return ParamObj(getattr, self, name)
        else:
            raise AttributeError(f"No attribute named '{name}' in {self}.")

    def __str__(self) -> str:
        args = [str(a) for a in self.args]
        kwargs = [f"{key}={str(value)}" for key, value in self.kwargs.items()]
        if isinstance(self.cls, Parametrized):
            name = str(self.cls)
        elif (
            hasattr(self.args[0], self.cls.__name__)
            and inspect.isfunction(self.cls)
            and inspect.isclass(self.args[0])
        ):
            name = f"{self.args[0].__name__}.{self.cls.__name__}"
            args = args[1:]
        else:
            name = self.cls.__name__
        return f"{name}({', '.join(args+kwargs)})"



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from unittest.mock import patch

import numpy as np
import pytest

import pulser
from pulser import Sequence, Pulse, Register
from pulser.devices import Chadoq2, MockDevice
from pulser.devices._device_datacls import Device
from pulser.sequence import _TimeSlot
from pulser.waveforms import (
    BlackmanWaveform,
    CompositeWaveform,
    RampWaveform,
    InterpolatedWaveform,
)

reg = Register.triangular_lattice(4, 7, spacing=5, prefix="q")
device = Chadoq2


def test_init():
    with pytest.raises(TypeError, match="must be of type 'Device'"):
        Sequence(reg, Device)

    fake_device = Device("fake", 2, 100, 100, 1, Chadoq2._channels)
    with pytest.warns(UserWarning, match="imported from 'pulser.devices'"):
        Sequence(reg, fake_device)

    seq = Sequence(reg, device)
    assert seq.qubit_info == reg.qubits
    assert seq.declared_channels == {}
    assert seq.available_channels.keys() == device.channels.keys()


def test_channel_declaration():
    seq = Sequence(reg, device)
    available_channels = set(seq.available_channels)
    seq.declare_channel("ch0", "rydberg_global")
    seq.declare_channel("ch1", "raman_local")
    with pytest.raises(ValueError, match="No channel"):
        seq.declare_channel("ch2", "raman")
    with pytest.raises(ValueError, match="not available"):
        seq.declare_channel("ch2", "rydberg_global")
    with pytest.raises(ValueError, match="name is already in use"):
        seq.declare_channel("ch0", "raman_local")

    chs = {"rydberg_global", "raman_local"}
    assert seq._schedule["ch0"][-1] == _TimeSlot(
        "target", -1, 0, set(seq.qubit_info.keys())
    )
    assert set(seq.available_channels) == available_channels - chs

    seq2 = Sequence(reg, MockDevice)
    available_channels = set(seq2.available_channels)
    seq2.declare_channel("ch0", "raman_local", initial_target="q1")
    seq2.declare_channel("ch1", "rydberg_global")
    seq2.declare_channel("ch2", "rydberg_global")
    assert set(seq2.available_channels) == available_channels - {"mw_global"}
    assert seq2._taken_channels == {
        "ch0": "raman_local",
        "ch1": "rydberg_global",
        "ch2": "rydberg_global",
    }
    assert seq2._taken_channels.keys() == seq2._channels.keys()
    with pytest.raises(ValueError, match="type 'Microwave' cannot work "):
        seq2.declare_channel("ch3", "mw_global")

    seq2 = Sequence(reg, MockDevice)
    seq2.declare_channel("ch0", "mw_global")
    assert set(seq2.available_channels) == {"mw_global"}
    with pytest.raises(
        ValueError,
        match="cannot work simultaneously with the declared 'Microwave'",
    ):
        seq2.declare_channel("ch3", "rydberg_global")


def test_target():
    seq = Sequence(reg, device)
    seq.declare_channel("ch0", "raman_local", initial_target="q1")
    seq.declare_channel("ch1", "rydberg_global")

    with pytest.raises(ValueError, match="name of a declared channel"):
        seq.target("q0", "ch2")
    with pytest.raises(ValueError, match="qubits must belong"):
        seq.target(0, "ch0")
        seq.target("0", "ch1")
    with pytest.raises(ValueError, match="Can only choose target of 'Local'"):
        seq.target("q3", "ch1")
    with pytest.raises(ValueError, match="can target at most 1 qubits"):
        seq.target(["q1", "q5"], "ch0")

    assert seq._schedule["ch0"][-1] == _TimeSlot("target", -1, 0, {"q1"})
    seq.target("q4", "ch0")
    retarget_t = seq.declared_channels["ch0"].retarget_time
    assert seq._schedule["ch0"][-1] == _TimeSlot(
        "target", 0, retarget_t, {"q4"}
    )
    seq.target("q4", "ch0")  # targets the same qubit
    seq.target("q20", "ch0")
    assert seq._schedule["ch0"][-1] == _TimeSlot(
        "target", retarget_t, 2 * retarget_t, {"q20"}
    )
    seq.delay(216, "ch0")
    seq.target("q2", "ch0")
    ti = 2 * retarget_t + 216
    tf = ti + 16
    assert seq._schedule["ch0"][-1] == _TimeSlot("target", ti, tf, {"q2"})

    seq.delay(220, "ch0")
    seq.target("q1", "ch0")
    ti = tf + 220
    tf = ti
    assert seq._schedule["ch0"][-1] == _TimeSlot("target", ti, tf, {"q1"})

    seq.delay(100, "ch0")
    seq.target("q10", "ch0")
    ti = tf + 100
    tf = ti + 120
    assert seq._schedule["ch0"][-1] == _TimeSlot("target", ti, tf, {"q10"})

    seq2 = Sequence(reg, MockDevice)
    seq2.declare_channel("ch0", "raman_local", initial_target={"q1", "q10"})
    seq2.phase_shift(1, "q2")
    with pytest.raises(ValueError, match="qubits with different phase"):
        seq2.target({"q3", "q1", "q2"}, "ch0")


def test_delay():
    seq = Sequence(reg, device)
    seq.declare_channel("ch0", "raman_local")
    with pytest.raises(ValueError, match="Use the name of a declared channel"):
        seq.delay(1e3, "ch01")
    with pytest.raises(ValueError, match="channel has no target"):
        seq.delay(100, "ch0")
    seq.target("q19", "ch0")
    seq.delay(388, "ch0")
    assert seq._last("ch0") == _TimeSlot("delay", 0, 388, {"q19"})


def test_delay_min_duration():
    # Check that a delay shorter than a channel's minimal duration
    # is automatically extended to that minimal duration
    seq = Sequence(reg, device)
    seq.declare_channel("ch0", "rydberg_global")
    seq.declare_channel("ch1", "rydberg_local")
    seq.target("q0", "ch1")
    pulse0 = Pulse.ConstantPulse(52, 1, 1, 0)
    pulse1 = Pulse.ConstantPulse(180, 1, 1, 0)
    seq.add(pulse1, "ch1")
    seq.add(pulse0, "ch0")
    seq.target("q1", "ch1")
    seq.add(pulse1, "ch1")
    min_duration = seq._channels["ch1"].min_duration
    assert seq._schedule["ch1"][3] == _TimeSlot(
        "delay", 220, 220 + min_duration, {"q1"}
    )


def test_min_pulse_duration():
    seq = Sequence(reg, device)
    seq.declare_channel("ch0", "rydberg_global")
    seq.declare_channel("ch1", "rydberg_local")
    seq.target("q0", "ch1")
    pulse0 = Pulse.ConstantPulse(60, 1, 1, 0)
    pulse1 = Pulse.ConstantPulse(80, 1, 1, 0)
    seq.add(pulse1, "ch1")
    assert seq._min_pulse_duration() == 80
    seq.add(pulse0, "ch0")
    seq.delay(52, "ch0")
    seq.target("q1", "ch1")
    seq.add(pulse1, "ch1")
    assert seq._min_pulse_duration() == 60


def test_phase():
    seq = Sequence(reg, device)
    seq.declare_channel("ch0", "raman_local", initial_target="q0")
    seq.phase_shift(-1, "q0", "q1")
    with pytest.raises(ValueError, match="id of a qubit declared"):
        seq.current_phase_ref(0, "digital")
    with pytest.raises(ValueError, match="targets the given 'basis'"):
        seq.current_phase_ref("q1", "ground-rydberg")
    with pytest.raises(ValueError, match="No declared channel targets"):
        seq.phase_shift(1, "q3", basis="hyperfine")
    assert seq.current_phase_ref("q0", "digital") == 2 * np.pi - 1

    # Phase shifts of 0
    seq.phase_shift(0, "q0")
    seq.phase_shift(-8 * np.pi, "q1")
    assert seq.current_phase_ref("q0", "digital") == 2 * np.pi - 1
    assert seq.current_phase_ref("q1", "digital") == 2 * np.pi - 1

    with pytest.raises(ValueError, match="targets have to be qubit ids"):
        seq.phase_shift(np.pi, "q1", "q4", "q100")

    seq.declare_channel("ch1", "rydberg_global")
    seq.phase_shift(1, *seq._qids, basis="ground-rydberg")
    for q in seq.qubit_info:
        assert seq.current_phase_ref(q, "ground-rydberg") == 1
    seq.phase_shift(1, *seq._qids)
    assert seq.current_phase_ref("q1", "digital") == 0
    assert seq.current_phase_ref("q10", "digital") == 1


def test_align():
    seq = Sequence(reg, device)
    seq.declare_channel("ch0", "raman_local", initial_target="q0")
    seq.declare_channel("ch1", "rydberg_global")
    with pytest.raises(ValueError, match="names must correspond to declared"):
        seq.align("ch0", "ch1", "ch2")
    with pytest.raises(ValueError, match="more than once"):
        seq.align("ch0", "ch1", "ch0")
    with pytest.raises(ValueError, match="at least two channels"):
        seq.align()
        seq.align("ch1")


def test_measure():
    pulse = Pulse.ConstantPulse(500, 2, -10, 0, post_phase_shift=np.pi)
    seq = Sequence(reg, MockDevice)
    seq.declare_channel("ch0", "rydberg_global")
    assert "XY" in MockDevice.supported_bases
    with pytest.raises(ValueError, match="not supported"):
        seq.measure(basis="XY")
    seq.measure()
    with pytest.raises(RuntimeError, match="already been measured"):
        seq.measure(basis="digital")
    with pytest.raises(RuntimeError, match="Nothing more can be added."):
        seq.add(pulse, "ch0")

    seq = Sequence(reg, MockDevice)
    seq.declare_channel("ch0", "mw_global")
    assert "digital" in MockDevice.supported_bases
    with pytest.raises(ValueError, match="not supported"):
        seq.measure(basis="digital")
    seq.measure(basis="XY")


def test_str():
    seq = Sequence(reg, device)
    seq.declare_channel("ch0", "raman_local", initial_target="q0")
    pulse = Pulse.ConstantPulse(500, 2, -10, 0, post_phase_shift=np.pi)
    seq.add(pulse, "ch0")
    seq.delay(200, "ch0")
    seq.target("q7", "ch0")
    seq.measure("digital")
    msg = (
        "Channel: ch0\nt: 0 | Initial targets: q0 | Phase Reference: 0.0 "
        + "\nt: 0->500 | Pulse(Amp=2 rad/µs, Detuning=-10 rad/µs, Phase=0) "
        + "| Targets: q0\nt: 500->700 | Delay \nt: 700->700 | Target: q7 | "
        + "Phase Reference: 0.0\n\nMeasured in basis: digital"
    )
    assert seq.__str__() == msg


def test_sequence():
    seq = Sequence(reg, device)
    assert seq.get_duration() == 0
    with pytest.raises(RuntimeError, match="empty sequence"):
        seq.draw()
    seq.declare_channel("ch0", "raman_local", initial_target="q0")
    seq.declare_channel("ch1", "rydberg_local", initial_target="q0")
    seq.declare_channel("ch2", "rydberg_global")
    assert seq.get_duration("ch0") == 0
    assert seq.get_duration("ch2") == 0
    seq.phase_shift(np.pi, "q0", basis="ground-rydberg")

    with patch("matplotlib.pyplot.show"):
        seq.draw()

    pulse1 = Pulse(
        InterpolatedWaveform(500, [0, 1, 0]),
        InterpolatedWaveform(500, [-1, 1, 0]),
        phase=0,
        post_phase_shift=np.pi,
    )
    pulse2 = Pulse.ConstantDetuning(
        BlackmanWaveform(1e3, np.pi / 4), 25, np.pi, post_phase_shift=1
    )
    with pytest.raises(TypeError):
        seq.add([1, 5, 3], "ch0")
    with pytest.raises(ValueError, match="amplitude goes over the maximum"):
        seq.add(
            Pulse.ConstantPulse(20, 2 * np.pi * 10, -2 * np.pi * 100, 0), "ch2"
        )
    with pytest.raises(
        ValueError, match="detuning values go out of the range"
    ):
        seq.add(
            Pulse.ConstantPulse(500, 2 * np.pi, -2 * np.pi * 100, 0), "ch0"
        )
    with pytest.raises(ValueError, match="qubits with different phase ref"):
        seq.add(pulse2, "ch2")
    with pytest.raises(ValueError, match="Invalid protocol"):
        seq.add(pulse1, "ch0", protocol="now")

    wf_ = CompositeWaveform(BlackmanWaveform(30, 1), RampWaveform(15, 0, 2))
    with pytest.raises(TypeError, match="Failed to automatically adjust"):
        with pytest.warns(UserWarning, match="rounded up to 48 ns"):
            seq.add(Pulse.ConstantAmplitude(1, wf_, 0), "ch0")

    pulse1_ = Pulse.ConstantPulse(499, 2, -10, 0, post_phase_shift=np.pi)
    with pytest.warns(UserWarning, match="rounded up to 500 ns"):
        seq.add(pulse1_, "ch0")
    seq.add(pulse1, "ch1")
    seq.add(pulse2, "ch2")

    assert seq._last("ch0").ti == 0
    assert seq._last("ch0").tf == seq._last("ch1").ti
    assert seq._last("ch2").tf == seq._last("ch2").ti + 1000
    assert seq.current_phase_ref("q0", "digital") == np.pi

    seq.add(pulse1, "ch2")
    assert seq.get_duration("ch2") == 2500
    seq.add(pulse2, "ch1", protocol="no-delay")
    assert seq.get_duration("ch1") == 3500
    seq.add(pulse1, "ch0", protocol="no-delay")
    assert seq._last("ch0").ti == 500
    assert seq.get_duration("ch0") == 1000
    assert seq.current_phase_ref("q0", "digital") == 0
    seq.phase_shift(np.pi / 2, "q1")
    seq.target("q1", "ch0")
    assert seq._last_used["digital"]["q1"] == 0
    assert seq._last_target["ch0"] == 1000
    assert seq._last("ch0").ti == 1000
    assert seq.get_duration("ch0") == 1000
    seq.add(pulse1, "ch0")
    assert seq._last("ch0").ti == 2500
    assert seq.get_duration("ch0") == 3000
    seq.add(pulse1, "ch0", protocol="wait-for-all")
    assert seq._last("ch0").ti == 3500
    assert seq.get_duration("ch2") != seq.get_duration("ch0")
    seq.align("ch0", "ch2")
    assert seq.get_duration("ch2") == seq.get_duration("ch0")

    with patch("matplotlib.pyplot.show"):
        seq.draw(draw_phase_shifts=True)

    assert seq.get_duration() == 4000

    seq.measure(basis="digital")

    with patch("matplotlib.pyplot.show"):
        seq.draw(draw_phase_area=True)

    s = seq.serialize()
    assert json.loads(s)["__version__"] == pulser.__version__
    seq_ = Sequence.deserialize(s)
    assert str(seq) == str(seq_)



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for JSON serializations."""

from __future__ import annotations

from typing import Any, Optional


def obj_to_dict(
    obj: object,
    *args: Any,
    _build: bool = True,
    _module: Optional[str] = None,
    _name: Optional[str] = None,
    _submodule: Optional[str] = None,
    **kwargs: Any
) -> dict[str, Any]:
    """Encodes an object in a dictionary for serialization.

    Args:
        obj: The object to encode in the dictionary.

    Other Parameters:
        _build (bool): Whether the object is to be built on deserialization.
        _module (str): Custom name for the module containing the object.
        _name (str): Custom name of the object.
        _submodule(str): Name of a submodule (e.g. the class holding a
                         classmethod). Only used when defined.
        args: If the object is to be built, the arguments to give on creation.
        kwargs: If the object is to be built, the keyword arguments to give on
            creation.

    Returns:
        dict: The dictionary encoding the object.
    """
    d = {
        "_build": _build,
        "__module__": _module if _module else obj.__class__.__module__,
        "__name__": _name if _name else obj.__class__.__name__,
    }
    if _build:
        d["__args__"] = args
        d["__kwargs__"] = kwargs
    if _submodule:
        d["__submodule__"] = _submodule

    return d



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes for containing and processing the results of a simulation."""

from __future__ import annotations

from collections import Counter
import collections.abc
from abc import ABC, abstractmethod
from functools import lru_cache
from typing import Optional, Union, cast, Tuple, Mapping

import matplotlib.pyplot as plt
import qutip
from qutip.piqs import isdiagonal
import numpy as np
from numpy.typing import ArrayLike


class SimulationResults(ABC):
    """Results of a simulation run of a pulse sequence.

    Parent class for NoisyResults and CoherentResults.
    Contains methods for studying the states and extracting useful information
    from them.
    """

    def __init__(
        self, size: int, basis_name: str, sim_times: np.ndarray
    ) -> None:
        """Initializes a new SimulationResults instance.

        Args:
            size (int): The number of atoms in the register.
            basis_name (str): The basis indicating the addressed atoms after
                the pulse sequence ('ground-rydberg', 'digital' or 'all').
            sim_times (array): Array of times (µs) when simulation results are
                returned.
        """
        self._dim = 3 if basis_name == "all" else 2
        self._size = size
        if basis_name not in {"ground-rydberg", "digital", "all"}:
            raise ValueError(
                "`basis_name` must be 'ground-rydberg', 'digital' or 'all'."
            )
        self._basis_name = basis_name
        self._sim_times = sim_times
        self._results: Union[list[Counter], list[qutip.Qobj]]
        # Use the pseudo-density matrix when calculating expectation values
        self._use_pseudo_dens: bool = False

    @property
    @abstractmethod
    def states(self) -> list[qutip.Qobj]:
        """Lists states of the system at simulation times."""
        pass

    @abstractmethod
    def get_state(self, t: float) -> qutip.Qobj:
        """Returns the state of the system at time t."""
        pass

    @abstractmethod
    def get_final_state(self) -> qutip.Qobj:
        """Returns the final state of the system."""
        pass

    @abstractmethod
    def _calc_weights(self, t_index: int) -> np.ndarray:
        """Computes the bitstring probabilities for sampled states."""
        pass

    def expect(
        self, obs_list: collections.abc.Sequence[Union[qutip.Qobj, ArrayLike]]
    ) -> list[Union[float, complex, ArrayLike]]:
        """Returns the expectation values of operators in obs_list.

        Args:
            obs_list (list[Union[qutip.Qobj, ArrayLike]]): Input observable
                list. ArrayLike objects will be converted to qutip.Qobj.

        Returns:
            list[Union[float, complex, ArrayLike]]: Expectation values of
            obs_list.
        """
        if not isinstance(obs_list, (list, np.ndarray)):
            raise TypeError("`obs_list` must be a list of operators.")

        qobj_list = []
        dim = self._dim if not self._use_pseudo_dens else 2
        legal_dims = [[dim] * self._size] * 2
        legal_shape = (dim ** self._size, dim ** self._size)
        for obs in obs_list:
            if not (
                isinstance(obs, np.ndarray) or isinstance(obs, qutip.Qobj)
            ):
                raise TypeError(
                    f"Incompatible type {type(obs)} of "
                    + "observable. Type must be ArrayLike or "
                    + "qutip.Qobj."
                )
            if obs.shape != legal_shape:
                raise ValueError(
                    "Incompatible shape of observable."
                    + f"Expected {legal_shape}, got {obs.shape}."
                )
            qobj_list.append(qutip.Qobj(obs, dims=legal_dims))
            if self._use_pseudo_dens:
                if not isdiagonal(obs):
                    raise ValueError(f"Observable {obs!r} is non-diagonal.")
                states = [
                    self._calc_pseudo_density(ind)
                    for ind in range(len(self._results))
                ]
            else:
                states = self.states

        return cast(list, qutip.expect(qobj_list, states))

    def sample_state(
        self, t: float, n_samples: int = 1000, t_tol: float = 1.0e-3
    ) -> Counter:
        """Returns the result of multiple measurements at time t.

        Args:
            t (float): Time at which the state is sampled.
            n_samples (int): Number of samples to return.
            t_tol (float): Tolerance for the difference between t and
                closest time.

        Returns:
            Counter: Sample distribution of bitstrings corresponding to
            measured quantum states at time t.
        """
        t_index = self._get_index_from_time(t, t_tol)
        dist = np.random.multinomial(n_samples, self._calc_weights(t_index))
        return Counter(
            {
                np.binary_repr(i, self._size): dist[i]
                for i in np.nonzero(dist)[0]
            }
        )

    def sample_final_state(self, N_samples: int = 1000) -> Counter:
        """Returns the result of multiple measurements of the final state.

        Args:
            N_samples (int): Number of samples to return.

        Returns:
            Counter: Sample distribution of bitstrings corresponding to
            measured quantum states at the end of the simulation.
        """
        return self.sample_state(self._sim_times[-1], N_samples)

    def plot(self, op: qutip.Qobj, fmt: str = "", label: str = "") -> None:
        """Plots the expectation value of a given operator op.

        Args:
            op (qutip.Qobj): Operator whose expectation value is wanted.
            fmt (str): Curve plot format.
            label (str): Curve label.
        """
        plt.plot(self._sim_times, self.expect([op])[0], fmt, label=label)
        plt.xlabel("Time (µs)")
        plt.ylabel("Expectation value")

    def _get_index_from_time(self, t_float: float, tol: float = 1.0e-3) -> int:
        """Returns closest index corresponding to time t_float.

        Args:
            t_float (float): Time value (in µs).
            tol (float): Tolerance for the difference between t_float and
                closest time.
        """
        try:
            return int(np.where(abs(t_float - self._sim_times) < tol)[0][0])
        except IndexError:
            raise IndexError(
                f"Given time {t_float} is absent from Simulation times within"
                + f" tolerance {tol}."
            )

    @lru_cache(maxsize=None)
    def _calc_pseudo_density(self, t_index: int) -> qutip.Qobj:
        """Calculates the pseudo-density matrix at a given time.

        The pseudo-density matrix is the diagonal matrix calculated from the
        probability of obtaining each possible state, after measurement.

        Args:
            t_index (int): The index in the list of states/results to turn
                into the pseudo-density matrix.

        Returns:
            qutip.Qobj: The pseudo-density matrix as a Qobj.
        """

        def _proj_from_bitstring(bitstring: str) -> qutip.Qobj:
            proj = qutip.tensor(
                [self._meas_projector(int(i)) for i in bitstring]
            )
            return proj

        w = self._calc_weights(t_index)
        return sum(
            w[i] * _proj_from_bitstring(np.binary_repr(i, width=self._size))
            for i in np.nonzero(w)[0]
        )

    def _meas_projector(self, state_n: int) -> qutip.Qobj:
        """Gets the post measurement projector (defaults to the ideal case).

        Args:
            state_n: The measured state (0 or 1).
        """
        if self._basis_name == "ground-rydberg":
            # 0 = |g> = |1>, 1 = |r> = |0>
            return qutip.basis(2, 1 - state_n).proj()

        return qutip.basis(2, state_n).proj()


class NoisyResults(SimulationResults):
    """Results of a noisy simulation run of a pulse sequence.

    Contrary to a CoherentResults object, this object contains a list of
    Counters describing the state distribution at the time it was created by
    using Simulation.run() with a noisy simulation.
    Contains methods for studying the populations and extracting useful
    information from them.
    """

    def __init__(
        self,
        run_output: list[Counter],
        size: int,
        basis_name: str,
        sim_times: np.ndarray,
        n_measures: int,
    ) -> None:
        """Initializes a new NoisyResults instance.

        Warning:
            Can't have single-atom Hilbert spaces with dimension bigger
            than 2 for NoisyResults objects.
            This is not the case for a CoherentResults object, containing
            states in Hilbert space, but NoisyResults contains a probability
            distribution of bitstrings, not atomic states

        Args:
            run_output (list[Counter]): Each Counter contains the
                probability distribution of a multi-qubits state,
                represented as a bitstring. There is one Counter for each time
                the simulation was asked to return a result.
            size (int): The number of atoms in the register.
            basis_name (str): Basis indicating the addressed atoms after
                the pulse sequence ('ground-rydberg' or 'digital' - 'all' basis
                makes no sense after projection on bitstrings). Defaults to
                'digital' if given value 'all'.
            sim_times (np.ndarray): Times at which Simulation object returned
                the results.
            meas_basis (Optional[str]): The basis in which a sampling
                measurement is desired.
            n_measures (int): Number of measurements needed to compute this
                result when doing the simulation.
        """
        basis_name_ = "digital" if basis_name == "all" else basis_name
        super().__init__(size, basis_name_, sim_times)
        self.n_measures = n_measures
        self._results = run_output
        self._use_pseudo_dens = True

    @property
    def states(self) -> list[qutip.Qobj]:
        """Measured states as a list of diagonal qutip.Qobj."""
        return [self.get_state(t) for t in self._sim_times]

    @property
    def results(self) -> list[Counter]:
        """Probability distribution of the bitstrings."""
        return self._results

    def get_state(self, t: float, t_tol: float = 1.0e-3) -> qutip.Qobj:
        """Gets the state at time t as a diagonal density matrix.

        Note:
            This is not the density matrix of the system, but is a convenient
            way of computing expectation values of observables.

        Args:
            t (float): Time (µs) at which to return the state.
            t_tol (float): Tolerance for the difference between t and
                closest time.

        Returns:
            qutip.Qobj: States probability distribution as a diagonal
            density matrix.
        """
        t_index = self._get_index_from_time(t, t_tol)
        return self._calc_pseudo_density(t_index)

    def get_final_state(self) -> qutip.Qobj:
        """Get the final state of the simulation as a diagonal density matrix.

        Note:
            This is not the density matrix of the system, but is a convenient
            way of computing expectation values of observables.

        Returns:
            qutip.Qobj: States probability distribution as a density matrix.
        """
        return self.get_state(self._sim_times[-1])

    def _calc_weights(self, t_index: int) -> np.ndarray:
        weights = np.zeros(2 ** self._size)
        for bin_rep, prob in self._results[t_index].items():
            weights[int(bin_rep, base=2)] = prob
        return weights

    def plot(
        self,
        op: qutip.Qobj,
        fmt: str = ".",
        label: str = "",
        error_bars: bool = True,
    ) -> None:
        """Plots the expectation value of a given operator op.

        Note:
            The observable must be diagonal.

        Args:
            op (qutip.Qobj): Operator whose expectation value is wanted.
            fmt (str): Curve plot format.
            label (str): y-Axis label.
            error_bars (bool): Choose to display error bars.
        """

        def get_error_bars() -> Tuple[ArrayLike, ArrayLike]:
            moy = self.expect([op])[0]
            standard_dev = cast(
                np.ndarray,
                np.sqrt(qutip.variance(op, self.states) / self.n_measures),
            )
            return moy, standard_dev

        if error_bars:
            moy, st = get_error_bars()
            plt.errorbar(
                self._sim_times, moy, st, fmt=fmt, lw=1, capsize=3, label=label
            )
            plt.xlabel("Time (µs)")
            plt.ylabel("Expectation value")
        else:
            super().plot(op, fmt, label)


class CoherentResults(SimulationResults):
    """Results of a coherent simulation run of a pulse sequence.

    Contains methods for studying the states and extracting useful information
    from them.
    """

    def __init__(
        self,
        run_output: list[qutip.Qobj],
        size: int,
        basis_name: str,
        sim_times: np.ndarray,
        meas_basis: str,
        meas_errors: Optional[Mapping[str, float]] = None,
    ) -> None:
        """Initializes a new CoherentResults instance.

        Args:
            run_output (list of qutip.Qobj): List of `qutip.Qobj` corresponding
                to the states at each time step after the evolution has been
                simulated.
            size (int): The number of atoms in the register.
            basis_name (str): The basis indicating the addressed atoms after
                the pulse sequence ('ground-rydberg', 'digital' or 'all').
            sim_times (list): Times at which Simulation object returned the
                results.
            meas_basis (str): The basis in which a sampling measurement
                is desired.
            meas_errors (Optional[Mapping[str, float]]): If measurement errors
                are involved, give them in a dictionary with "epsilon" and
                "epsilon_prime".
        """
        super().__init__(size, basis_name, sim_times)
        if meas_basis:
            if meas_basis not in {"ground-rydberg", "digital"}:
                raise ValueError(
                    "`meas_basis` must be 'ground-rydberg' or 'digital'."
                )
        self._meas_basis = meas_basis
        self._results = run_output
        if meas_errors is not None:
            if set(meas_errors) != {"epsilon", "epsilon_prime"}:
                raise ValueError(
                    "When defining measurement errors, only values of "
                    "'epsilon' and 'epsilon_prime' must be given."
                )
            self._use_pseudo_dens = True
        self._meas_errors = meas_errors

    @property
    def states(self) -> list[qutip.Qobj]:
        """List of ``qutip.Qobj`` for each state in the simulation."""
        return list(self._results)

    def get_state(
        self,
        t: float,
        reduce_to_basis: Optional[str] = None,
        ignore_global_phase: bool = True,
        tol: float = 1e-6,
        normalize: bool = True,
        t_tol: float = 1.0e-3,
    ) -> qutip.Qobj:
        """Get the state at time t of the simulation.

        Args:
            t (float): Time (µs) at which to return the state.
            reduce_to_basis (str, default=None): Reduces the full state vector
                to the given basis ("ground-rydberg" or "digital"), if the
                population of the states to be ignored is negligible.
            ignore_global_phase (bool, default=True): If True, changes the
                final state's global phase such that the largest term (in
                absolute value) is real.
            tol (float, default=1e-6): Maximum allowed population of each
                eliminated state.
            normalize (bool, default=True): Whether to normalize the reduced
                state.
            t_tol (float): Tolerance for the difference between t and
                closest time.

        Returns:
            qutip.Qobj: The resulting state at time t.

        Raises:
            TypeError: If trying to reduce to a basis that would eliminate
                states with significant occupation probabilites.
        """
        t_index = self._get_index_from_time(t, t_tol)
        state = cast(qutip.Qobj, self._results[t_index].copy())
        if ignore_global_phase:
            full = state.full()
            global_ph = float(np.angle(full[np.argmax(np.abs(full))]))
            state *= np.exp(-1j * global_ph)
        if self._dim != 3:
            if reduce_to_basis not in [None, self._basis_name]:
                raise TypeError(
                    f"Can't reduce a system in {self._basis_name}"
                    + f" to the {reduce_to_basis} basis."
                )
        elif reduce_to_basis is not None:
            if reduce_to_basis == "ground-rydberg":
                ex_state = "2"
            elif reduce_to_basis == "digital":
                ex_state = "0"
            else:
                raise ValueError(
                    "'reduce_to_basis' must be 'ground-rydberg' "
                    + f"or 'digital', not '{reduce_to_basis}'."
                )
            ex_inds = [
                i
                for i in range(3 ** self._size)
                if ex_state in np.base_repr(i, base=3).zfill(self._size)
            ]
            ex_probs = np.abs(state.extract_states(ex_inds).full()) ** 2
            if not np.all(np.isclose(ex_probs, 0, atol=tol)):
                raise TypeError(
                    "Can't reduce to chosen basis because the population of a "
                    "state to eliminate is above the allowed tolerance."
                )
            state = state.eliminate_states(ex_inds, normalize=normalize)
        return state.tidyup()

    def get_final_state(
        self,
        reduce_to_basis: Optional[str] = None,
        ignore_global_phase: bool = True,
        tol: float = 1e-6,
        normalize: bool = True,
    ) -> qutip.Qobj:
        """Returns the final state of the Simulation.

        Args:
            reduce_to_basis (str, default=None): Reduces the full state vector
                to the given basis ("ground-rydberg" or "digital"), if the
                population of the states to be ignored is negligible.
            ignore_global_phase (bool, default=True): If True, changes the
                final state's global phase such that the largest term (in
                absolute value) is real.
            tol (float, default=1e-6): Maximum allowed population of each
                eliminated state.
            normalize (bool, default=True): Whether to normalize the reduced
                state.

        Returns:
            qutip.Qobj: The resulting final state.

        Raises:
            TypeError: If trying to reduce to a basis that would eliminate
                states with significant occupation probabilites.
        """
        return self.get_state(
            self._sim_times[-1],
            reduce_to_basis,
            ignore_global_phase,
            tol,
            normalize,
        )

    def _calc_weights(self, t_index: int) -> np.ndarray:
        n = self._size
        state_t = cast(qutip.Qobj, self._results[t_index]).unit()
        if state_t.type != "ket":
            probs = np.abs(state_t.diag())
        else:
            probs = (np.abs(state_t.full()) ** 2).flatten()

        if self._dim == 2:
            if self._meas_basis == self._basis_name:
                # State vector ordered with r first for 'ground_rydberg'
                # e.g. n=2: [rr, rg, gr, gg] -> [11, 10, 01, 00]
                # Invert the order ->  [00, 01, 10, 11] correspondence
                weights = (
                    probs if self._meas_basis == "digital" else probs[::-1]
                )
            else:
                # Only 000...000 is measured
                weights = np.zeros(probs.size)
                weights[0] = 1.0

        elif self._dim == 3:
            if self._meas_basis == "ground-rydberg":
                one_state = 0  # 1 = |r>
                ex_one = slice(1, 3)
            elif self._meas_basis == "digital":
                one_state = 2  # 1 = |h>
                ex_one = slice(0, 2)
            probs = probs.reshape([3] * n)
            weights = np.zeros(2 ** n)
            for dec_val in range(2 ** n):
                ind: list[Union[int, slice]] = []
                for v in np.binary_repr(dec_val, width=n):
                    if v == "0":
                        ind.append(ex_one)
                    else:
                        ind.append(one_state)
                # Eg: 'digital' basis : |1> = index2, |0> = index0, 1 = 0:2
                # p_11010 = sum(probs[2, 2, 0:2, 2, 0:2])
                # We sum all probabilites that correspond to measuring
                # 11010, namely hhghg, hhrhg, hhghr, hhrhr
                weights[dec_val] = np.sum(probs[tuple(ind)])
        else:
            raise NotImplementedError(
                "Cannot sample system with single-atom state vectors of "
                "dimension > 3."
            )
        # Takes care of numerical artefacts in case sum(weights) != 1
        weights /= sum(weights)
        return cast(np.ndarray, weights)

    def _meas_projector(self, state_n: int) -> qutip.Qobj:
        if self._meas_errors:
            err_param = (
                self._meas_errors["epsilon"]
                if state_n == 0
                else self._meas_errors["epsilon_prime"]
            )
            # 'good' is the position of the state that measures to state_n
            # Matches for the digital basis, is inverted for ground-rydberg
            good = state_n if self._basis_name == "digital" else 1 - state_n
            return (
                qutip.basis(2, good).proj() * (1 - err_param)
                + qutip.basis(2, 1 - good).proj() * err_param
            )
        # Returns normal projectors in the absence of measurement errors
        return super()._meas_projector(state_n)

    def sample_state(
        self, t: float, n_samples: int = 1000, t_tol: float = 1.0e-3
    ) -> Counter:
        """Returns the result of multiple measurements at time t.

        Args:
            t (float): Time at which the state is sampled.
            n_samples (int): Number of samples to return.
            t_tol (float): Tolerance for the difference between t and
                closest time.

        Returns:
            Counter: Sample distribution of bitstrings corresponding to
            measured quantum states at time t.
        """
        sampled_state = super().sample_state(t, n_samples, t_tol)
        if self._meas_errors is None:
            return sampled_state

        detected_sample_dict: Counter = Counter()
        for (shot, n_detects) in sampled_state.items():
            eps = self._meas_errors["epsilon"]
            eps_p = self._meas_errors["epsilon_prime"]
            # Shot as an array of 1s and 0s
            shot_arr = np.array(list(shot), dtype=int)
            # Probability of flipping each bit
            flip_probs = np.array([eps_p if x == "1" else eps for x in shot])
            # 1 if it flips, 0 if it stays the same
            flips = (
                np.random.uniform(size=(n_detects, len(flip_probs)))
                < flip_probs
            ).astype(int)
            # XOR betwen the original array and the flips
            # Gives an array of n_detects individual shots
            new_shots = shot_arr ^ flips
            # Count all the new_shots
            detected_sample_dict += Counter(
                "".join(map(str, measured)) for measured in new_shots
            )

        return detected_sample_dict



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains the abstract base class for parametrized objects."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from pulser.parametrized import Variable  # pragma: no cover


class Parametrized(ABC):
    """Abstract base class for a parametrized object."""

    @property
    @abstractmethod
    def variables(self) -> dict[str, Variable]:
        """All the variables involved with this object."""
        pass

    @abstractmethod
    def build(self) -> Any:
        """Builds the object."""
        pass

    @abstractmethod
    def _to_dict(self) -> dict[str, Any]:
        """Serializes the object in a dictionary."""
        pass



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains the Pulse class, the building block of a pulse sequence."""

from __future__ import annotations

import functools
import itertools
from typing import Any, cast, Union

import matplotlib.pyplot as plt
import numpy as np

from pulser.parametrized import Parametrized, ParamObj
from pulser.parametrized.decorators import parametrize
from pulser.waveforms import Waveform, ConstantWaveform
from pulser.json.utils import obj_to_dict


class Pulse:
    r"""A generic pulse.

    In Pulser, a Pulse is a modulation of a frequency signal in amplitude
    and/or frequency, with a specific phase, over a given duration. Amplitude
    and frequency modulations are defined by :class:`Waveform` child classes.
    Frequency modulation is determined by a detuning waveform, which describes
    the shift in frequency from the channel's central frequency over time.
    If either quantity is constant throughout the entire pulse, use the
    ``ConstantDetuning``, ``ConstantAmplitude`` or ``ConstantPulse`` class
    method to create it.

    Note:
        We define the ``amplitude`` of a pulse to be its Rabi frequency,
        :math:`\Omega`, in rad/µs. Equivalently, the ``detuning`` is
        :math:`\delta`, also in rad/µs.

    Args:
        amplitude (Waveform): The pulse amplitude waveform.
        detuning (Waveform): The pulse detuning waveform.
        phase (float): The pulse phase (in radians).
        post_phase_shift (float, default=0.): Optionally lets you add a phase
            shift(in rads) immediately after the end of the pulse. This allows
            for enconding of arbitrary single-qubit gates into a single pulse
            (see ``Sequence.phase_shift()`` for more information).
    """

    def __new__(cls, *args, **kwargs):  # type: ignore
        """Creates a Pulse instance or a ParamObj depending on the input."""
        for x in itertools.chain(args, kwargs.values()):
            if isinstance(x, Parametrized):
                return ParamObj(cls, *args, **kwargs)
        else:
            return object.__new__(cls)

    def __init__(
        self,
        amplitude: Union[Waveform, Parametrized],
        detuning: Union[Waveform, Parametrized],
        phase: Union[float, Parametrized],
        post_phase_shift: Union[float, Parametrized] = 0.0,
    ):
        """Initializes a new Pulse."""
        if not (
            isinstance(amplitude, Waveform) and isinstance(detuning, Waveform)
        ):
            raise TypeError("'amplitude' and 'detuning' have to be waveforms.")

        if detuning.duration != amplitude.duration:
            raise ValueError(
                "The duration of detuning and amplitude waveforms must match."
            )
        self.duration = amplitude.duration
        if np.any(amplitude.samples < 0):
            raise ValueError(
                "All samples of an amplitude waveform must be "
                "greater than or equal to zero."
            )
        self.amplitude = amplitude
        self.detuning = detuning
        phase = cast(float, phase)
        self.phase = float(phase) % (2 * np.pi)
        post_phase_shift = cast(float, post_phase_shift)
        self.post_phase_shift = float(post_phase_shift) % (2 * np.pi)

    @classmethod
    @parametrize
    def ConstantDetuning(
        cls,
        amplitude: Union[Waveform, Parametrized],
        detuning: Union[float, Parametrized],
        phase: Union[float, Parametrized],
        post_phase_shift: Union[float, Parametrized] = 0.0,
    ) -> Pulse:
        """Creates a Pulse with an amplitude waveform and a constant detuning.

        Args:
            amplitude (Waveform): The pulse amplitude waveform.
            detuning (float): The detuning value (in rad/µs).
            phase (float): The pulse phase (in radians).
            post_phase_shift (float, default=0.): Optionally lets you add a
                phase shift (in rads) immediately after the end of the pulse.
        """
        detuning_wf = ConstantWaveform(
            cast(Waveform, amplitude).duration, detuning
        )
        return cls(amplitude, detuning_wf, phase, post_phase_shift)

    @classmethod
    @parametrize
    def ConstantAmplitude(
        cls,
        amplitude: Union[float, Parametrized],
        detuning: Union[Waveform, Parametrized],
        phase: Union[float, Parametrized],
        post_phase_shift: Union[float, Parametrized] = 0.0,
    ) -> Pulse:
        """Pulse with a constant amplitude and a detuning waveform.

        Args:
            amplitude (float): The pulse amplitude value (in rad/µs).
            detuning (Waveform): The pulse detuning waveform.
            phase (float): The pulse phase (in radians).
            post_phase_shift (float, default=0.): Optionally lets you add a
                phase shift (in rads) immediately after the end of the pulse.
        """
        amplitude_wf = ConstantWaveform(
            cast(Waveform, detuning).duration, amplitude
        )
        return cls(amplitude_wf, detuning, phase, post_phase_shift)

    @classmethod
    @parametrize
    def ConstantPulse(
        cls,
        duration: Union[int, Parametrized],
        amplitude: Union[float, Parametrized],
        detuning: Union[float, Parametrized],
        phase: Union[float, Parametrized],
        post_phase_shift: Union[float, Parametrized] = 0.0,
    ) -> Pulse:
        """Pulse with a constant amplitude and a constant detuning.

        Args:
            duration (int): The pulse duration (in multiples of 4 ns).
            amplitude (float): The pulse amplitude value (in rad/µs).
            detuning (float): The detuning value (in rad/µs).
            phase (float): The pulse phase (in radians).
            post_phase_shift (float, default=0.): Optionally lets you add a
                phase shift (in rads) immediately after the end of the pulse.
        """
        amplitude_wf = ConstantWaveform(duration, amplitude)
        detuning_wf = ConstantWaveform(duration, detuning)
        return cls(amplitude_wf, detuning_wf, phase, post_phase_shift)

    def draw(self) -> None:
        """Draws the pulse's amplitude and frequency waveforms."""
        fig, ax1 = plt.subplots()
        ax2 = ax1.twinx()

        self.amplitude._plot(ax1, r"$\Omega$ (rad/µs)", color="darkgreen")
        self.detuning._plot(ax2, r"$\delta$ (rad/µs)", color="indigo")

        fig.tight_layout()
        plt.show()

    def _to_dict(self) -> dict[str, Any]:
        return obj_to_dict(
            self,
            self.amplitude,
            self.detuning,
            self.phase,
            post_phase_shift=self.post_phase_shift,
        )

    def __str__(self) -> str:
        return (
            f"Pulse(Amp={self.amplitude!s}, Detuning={self.detuning!s}, "
            f"Phase={self.phase:.3g})"
        )

    def __repr__(self) -> str:
        return (
            f"Pulse(amp={self.amplitude!r}, detuning={self.detuning!r}, "
            + f"phase={self.phase:.3g}, "
            + f"post_phase_shift={self.post_phase_shift:.3g})"
        )


# Replicate __init__'s signature in __new__
functools.update_wrapper(Pulse.__new__, Pulse.__init__)



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The various hardware channel types."""

from __future__ import annotations

from dataclasses import dataclass
from typing import cast, ClassVar, Optional
import warnings

# Warnings of adjusted waveform duration appear just once
warnings.filterwarnings("once", "A duration of")


@dataclass(init=True, repr=False, frozen=True)
class Channel:
    """Base class of a hardware channel.

    Not to be initialized itself, but rather through a child class and the
    ``Local`` or ``Global`` classmethods.

    Args:
        name: The name of channel.
        basis: The addressed basis name.
        addressing: "Local" or "Global".
        max_abs_detuning: Maximum possible detuning (in rad/µs), in absolute
            value.
        max_amp: Maximum pulse amplitude (in rad/µs).
        retarget_time: Maximum time to change the target (in ns).
        max_targets: How many qubits can be addressed at once by the same beam.
        clock_period: The duration of a clock cycle (in ns). The duration of a
            pulse or delay instruction is enforced to be a multiple of the
            clock cycle.
        min_duration: The shortest duration an instruction can take.
        max_duration: The longest duration an instruction can take.

    Example:
        To create a channel targeting the 'ground-rydberg' transition globally,
        call ``Rydberg.Global(...)``.
    """

    name: ClassVar[str]
    basis: ClassVar[str]
    addressing: str
    max_abs_detuning: float
    max_amp: float
    retarget_time: Optional[int] = None
    max_targets: Optional[int] = None
    clock_period: int = 4  # ns
    min_duration: int = 16  # ns
    max_duration: int = 67108864  # ns

    @classmethod
    def Local(
        cls,
        max_abs_detuning: float,
        max_amp: float,
        retarget_time: int = 220,
        max_targets: int = 1,
        **kwargs: int,
    ) -> Channel:
        """Initializes the channel with local addressing.

        Args:
            max_abs_detuning (float): Maximum possible detuning (in rad/µs), in
                absolute value.
            max_amp(float): Maximum pulse amplitude (in rad/µs).
            retarget_time (int): Maximum time to change the target (in ns).
            max_targets (int): Maximum number of atoms the channel can target
                simultaneously.
        """
        return cls(
            "Local",
            max_abs_detuning,
            max_amp,
            retarget_time,
            max_targets,
            **kwargs,
        )

    @classmethod
    def Global(
        cls, max_abs_detuning: float, max_amp: float, **kwargs: int
    ) -> Channel:
        """Initializes the channel with global addressing.

        Args:
            max_abs_detuning (float): Maximum possible detuning (in rad/µs), in
                absolute value.
            max_amp(float): Maximum pulse amplitude (in rad/µs).
        """
        return cls("Global", max_abs_detuning, max_amp, **kwargs)

    def validate_duration(self, duration: int) -> int:
        """Validates and adapts the duration of an instruction on this channel.

        Args:
            duration (int): The duration to validate.

        Returns:
            int: The duration, potentially adapted to the channels specs.
        """
        try:
            _duration = int(duration)
        except (TypeError, ValueError):
            raise TypeError(
                "duration needs to be castable to an int but "
                "type %s was provided" % type(duration)
            )

        if duration < self.min_duration:
            raise ValueError(
                "duration has to be at least " + f"{self.min_duration} ns."
            )

        if duration > self.max_duration:
            raise ValueError(
                "duration can be at most " + f"{self.max_duration} ns."
            )

        if duration % self.clock_period != 0:
            _duration += self.clock_period - _duration % self.clock_period
            warnings.warn(
                f"A duration of {duration} ns is not a multiple of "
                f"the channel's clock period ({self.clock_period} "
                f"ns). It was rounded up to {_duration} ns.",
                stacklevel=4,
            )
        return _duration

    def __repr__(self) -> str:
        config = (
            f".{self.addressing}(Max Absolute Detuning: "
            f"{self.max_abs_detuning} rad/µs, Max Amplitude: "
            f"{self.max_amp} rad/µs"
        )
        if self.addressing == "Local":
            config += f", Target time: {self.retarget_time} ns"
            if cast(int, self.max_targets) > 1:
                config += f", Max targets: {self.max_targets}"
        config += f", Basis: '{self.basis}'"
        return self.name + config + ")"


@dataclass(init=True, repr=False, frozen=True)
class Raman(Channel):
    """Raman beam channel.

    Channel targeting the transition between the hyperfine ground states, in
    which the 'digital' basis is encoded. See base class.
    """

    name: ClassVar[str] = "Raman"
    basis: ClassVar[str] = "digital"


@dataclass(init=True, repr=False, frozen=True)
class Rydberg(Channel):
    """Rydberg beam channel.

    Channel targeting the transition between the ground and rydberg states,
    thus enconding the 'ground-rydberg' basis. See base class.
    """

    name: ClassVar[str] = "Rydberg"
    basis: ClassVar[str] = "ground-rydberg"


@dataclass(init=True, repr=False, frozen=True)
class Microwave(Channel):
    """Microwave adressing channel.

    Channel targeting the transition between two rydberg states, thus encoding
    the 'XY' basis. See base class.
    """

    name: ClassVar[str] = "Microwave"
    basis: ClassVar[str] = "XY"



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import deepcopy

from collections import Counter

import numpy as np
import pytest
import qutip
from qutip.piqs import isdiagonal

from pulser import Sequence, Pulse, Register
from pulser.devices import Chadoq2
from pulser.waveforms import BlackmanWaveform
from pulser.simulation import Simulation, SimConfig
from pulser.simulation.simresults import CoherentResults, NoisyResults

np.random.seed(123)
q_dict = {
    "A": np.array([0.0, 0.0]),
    "B": np.array([0.0, 10.0]),
}
reg = Register(q_dict)

duration = 1000
pi = Pulse.ConstantDetuning(BlackmanWaveform(duration, np.pi), 0.0, 0)

seq = Sequence(reg, Chadoq2)

# Declare Channels
seq.declare_channel("ryd", "rydberg_global")
seq.add(pi, "ryd")
seq_no_meas = deepcopy(seq)
seq_no_meas_noisy = deepcopy(seq)
seq.measure("ground-rydberg")

sim = Simulation(seq)
cfg_noisy = SimConfig(noise=("SPAM", "doppler", "amplitude"))
sim_noisy = Simulation(seq, config=cfg_noisy)
results = sim.run()
results_noisy = sim_noisy.run()

state = qutip.tensor([qutip.basis(2, 0), qutip.basis(2, 0)])
ground = qutip.tensor([qutip.basis(2, 1), qutip.basis(2, 1)])


def test_initialization():
    with pytest.raises(ValueError, match="`basis_name` must be"):
        CoherentResults(state, 2, "bad_basis", None, [0])
    with pytest.raises(ValueError, match="`meas_basis` must be"):
        CoherentResults(
            state, 1, "ground-rydberg", [0], "wrong_measurement_basis"
        )
    with pytest.raises(ValueError, match="`basis_name` must be"):
        NoisyResults(state, 2, "bad_basis", [0], 123)
    with pytest.raises(
        ValueError, match="only values of 'epsilon' and 'epsilon_prime'"
    ):
        CoherentResults(
            state,
            1,
            "ground-rydberg",
            [0],
            "ground-rydberg",
            cfg_noisy.spam_dict,
        )

    assert results._dim == 2
    assert results._size == 2
    assert results._basis_name == "ground-rydberg"
    assert results._meas_basis == "ground-rydberg"
    assert results.states[0] == ground


def test_get_final_state():
    with pytest.raises(TypeError, match="Can't reduce"):
        results.get_final_state(reduce_to_basis="digital")
    assert (
        results.get_final_state(
            reduce_to_basis="ground-rydberg", ignore_global_phase=False
        )
        == results.states[-1].tidyup()
    )
    assert np.all(
        np.isclose(
            np.abs(results.get_final_state().full()),
            np.abs(results.states[-1].full()),
        )
    )

    seq_ = Sequence(reg, Chadoq2)
    seq_.declare_channel("ryd", "rydberg_global")
    seq_.declare_channel("ram", "raman_local", initial_target="A")
    seq_.add(pi, "ram")
    seq_.add(pi, "ram")
    seq_.add(pi, "ryd")

    sim_ = Simulation(seq_)
    results_ = sim_.run()

    with pytest.raises(ValueError, match="'reduce_to_basis' must be"):
        results_.get_final_state(reduce_to_basis="all")

    with pytest.raises(TypeError, match="Can't reduce to chosen basis"):
        results_.get_final_state(reduce_to_basis="digital")

    h_states = results_.get_final_state(
        reduce_to_basis="digital", tol=1, normalize=False
    ).eliminate_states([0])
    assert h_states.norm() < 3e-6

    assert np.all(
        np.isclose(
            np.abs(
                results_.get_final_state(
                    reduce_to_basis="ground-rydberg"
                ).full()
            ),
            np.abs(results.states[-1].full()),
            atol=1e-5,
        )
    )


def test_get_final_state_noisy():
    np.random.seed(123)
    seq_ = Sequence(reg, Chadoq2)
    seq_.declare_channel("ram", "raman_local", initial_target="A")
    seq_.add(pi, "ram")
    noisy_config = SimConfig(noise=("SPAM", "doppler"))
    sim_noisy = Simulation(seq_, config=noisy_config)
    res3 = sim_noisy.run()
    res3._meas_basis = "digital"
    final_state = res3.get_final_state()
    assert isdiagonal(final_state)
    res3._meas_basis = "ground-rydberg"
    assert (
        final_state[0, 0] == 0.06666666666666667 + 0j
        and final_state[2, 2] == 0.9333333333333333 + 0j
    )
    assert res3.states[-1] == final_state
    assert res3.results[-1] == Counter(
        {"10": 0.9333333333333333, "00": 0.06666666666666667}
    )


def test_get_state_float_time():
    with pytest.raises(IndexError, match="is absent from"):
        results.get_state(-1.0)
    with pytest.raises(IndexError, match="is absent from"):
        mean = (results._sim_times[-1] + results._sim_times[-2]) / 2
        diff = (results._sim_times[-1] - results._sim_times[-2]) / 2
        results.get_state(mean, t_tol=diff / 2)
    state = results.get_state(mean, t_tol=3 * diff / 2)
    assert state == results.get_state(results._sim_times[-2])
    assert np.isclose(
        state.full(),
        np.array(
            [
                [0.79602211 + 0.0j],
                [0.02417478 - 0.37829574j],
                [0.02417478 - 0.37829574j],
                [-0.27423657 - 0.06131009j],
            ]
        ),
    ).all()


def test_expect():
    with pytest.raises(TypeError, match="must be a list"):
        results.expect("bad_observable")
    with pytest.raises(TypeError, match="Incompatible type"):
        results.expect(["bad_observable"])
    with pytest.raises(ValueError, match="Incompatible shape"):
        results.expect([np.array(3)])
    reg_single = Register.from_coordinates([(0, 0)], prefix="q")
    seq_single = Sequence(reg_single, Chadoq2)
    seq_single.declare_channel("ryd", "rydberg_global")
    seq_single.add(pi, "ryd")
    sim_single = Simulation(seq_single)
    results_single = sim_single.run()
    op = [qutip.basis(2, 0).proj()]
    exp = results_single.expect(op)[0]
    assert np.isclose(exp[-1], 1)
    assert len(exp) == duration
    np.testing.assert_almost_equal(
        results_single._calc_pseudo_density(-1).full(),
        np.array([[1, 0], [0, 0]]),
    )

    config = SimConfig(noise="SPAM", eta=0)
    sim_single.set_config(config)
    sim_single.evaluation_times = "Minimal"
    results_single = sim_single.run()
    exp = results_single.expect(op)[0]
    assert len(exp) == 2
    assert isinstance(results_single, CoherentResults)
    assert results_single._meas_errors == {
        "epsilon": config.epsilon,
        "epsilon_prime": config.epsilon_prime,
    }
    # Probability of measuring 1 = probability of false positive
    assert np.isclose(exp[0], config.epsilon)
    # Probability of measuring 1 = 1 - probability of false negative
    assert np.isclose(exp[-1], 1 - config.epsilon_prime)
    np.testing.assert_almost_equal(
        results_single._calc_pseudo_density(-1).full(),
        np.array([[1 - config.epsilon_prime, 0], [0, config.epsilon_prime]]),
    )
    seq3dim = Sequence(reg, Chadoq2)
    seq3dim.declare_channel("ryd", "rydberg_global")
    seq3dim.declare_channel("ram", "raman_local", initial_target="A")
    seq3dim.add(pi, "ram")
    seq3dim.add(pi, "ryd")
    sim3dim = Simulation(seq3dim)
    exp3dim = sim3dim.run().expect(
        [qutip.tensor(qutip.basis(3, 0).proj(), qutip.qeye(3))]
    )
    assert np.isclose(exp3dim[0][-1], 1.89690200e-14)


def test_expect_noisy():
    np.random.seed(123)
    bad_op = qutip.tensor([qutip.qeye(2), qutip.sigmap()])
    with pytest.raises(ValueError, match="non-diagonal"):
        results_noisy.expect([bad_op])
    op = qutip.tensor([qutip.qeye(2), qutip.basis(2, 0).proj()])
    assert np.isclose(results_noisy.expect([op])[0][-1], 0.7733333333333334)


def test_plot():
    op = qutip.tensor([qutip.qeye(2), qutip.basis(2, 0).proj()])
    results_noisy.plot(op)
    results_noisy.plot(op, error_bars=False)
    results.plot(op)


def test_sample_final_state():
    np.random.seed(123)
    sim_no_meas = Simulation(seq_no_meas, config=SimConfig(runs=1))
    results_no_meas = sim_no_meas.run()
    assert results_no_meas.sample_final_state() == Counter(
        {"00": 77, "01": 140, "10": 167, "11": 616}
    )
    with pytest.raises(NotImplementedError, match="dimension > 3"):
        results_large_dim = deepcopy(results)
        results_large_dim._dim = 7
        results_large_dim.sample_final_state()

    sampling = results.sample_final_state(1234)
    assert len(sampling) == 4  # Check that all states were observed.

    results._meas_basis = "digital"
    sampling0 = results.sample_final_state(N_samples=911)
    assert sampling0 == {"00": 911}
    seq_no_meas.declare_channel("raman", "raman_local", "B")
    seq_no_meas.add(pi, "raman")
    res_3level = Simulation(seq_no_meas).run()
    # Raman pi pulse on one atom will not affect other,
    # even with global pi on rydberg
    assert len(res_3level.sample_final_state()) == 2
    res_3level._meas_basis = "ground-rydberg"
    sampling_three_levelB = res_3level.sample_final_state()
    # Rydberg will affect both:
    assert len(sampling_three_levelB) == 4


def test_sample_final_state_noisy():
    np.random.seed(123)
    assert results_noisy.sample_final_state(N_samples=1234) == Counter(
        {"11": 787, "10": 219, "01": 176, "00": 52}
    )
    res_3level = Simulation(
        seq_no_meas_noisy, config=SimConfig(noise=("SPAM", "doppler"), runs=10)
    )
    final_state = res_3level.run().states[-1]
    assert np.isclose(
        final_state.full(),
        np.array(
            [
                [0.62 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j],
                [0.0 + 0.0j, 0.16 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j],
                [0.0 + 0.0j, 0.0 + 0.0j, 0.18 + 0.0j, 0.0 + 0.0j],
                [0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j, 0.04 + 0.0j],
            ]
        ),
    ).all()



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections import defaultdict
from typing import Any, cast, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import CubicSpline

import pulser
from pulser.waveforms import ConstantWaveform, InterpolatedWaveform
from pulser.pulse import Pulse


def gather_data(seq: pulser.sequence.Sequence) -> dict:
    """Collects the whole sequence data for plotting.

    Args:
        seq (pulser.Sequence): The input sequence of operations on a device.

    Returns:
        dict: The data to plot.
    """
    # The minimum time axis length is 100 ns
    total_duration = max(seq.get_duration(), 100)
    data = {}
    for ch, sch in seq._schedule.items():
        time = [-1]  # To not break the "time[-1]" later on
        amp = []
        detuning = []
        # List of interpolation points
        interp_pts: defaultdict[str, list[list[float]]] = defaultdict(list)
        target: dict[Union[str, tuple[int, int]], Any] = {}
        # phase_shift = {}
        for slot in sch:
            if slot.ti == -1:
                target["initial"] = slot.targets
                time += [0]
                amp += [0.0]
                detuning += [0.0]
                continue
            if slot.type in ["delay", "target"]:
                time += [
                    slot.ti,
                    slot.tf - 1 if slot.tf > slot.ti else slot.ti,
                ]
                amp += [0.0, 0.0]
                detuning += [0.0, 0.0]
                if slot.type == "target":
                    target[(slot.ti, slot.tf - 1)] = slot.targets
                continue
            pulse = cast(Pulse, slot.type)
            if isinstance(pulse.amplitude, ConstantWaveform) and isinstance(
                pulse.detuning, ConstantWaveform
            ):
                time += [slot.ti, slot.tf - 1]
                amp += [float(pulse.amplitude._value)] * 2
                detuning += [float(pulse.detuning._value)] * 2
            else:
                time += list(range(slot.ti, slot.tf))
                amp += pulse.amplitude.samples.tolist()
                detuning += pulse.detuning.samples.tolist()
                for wf_type in ["amplitude", "detuning"]:
                    wf = getattr(pulse, wf_type)
                    if isinstance(wf, InterpolatedWaveform):
                        pts = wf.data_points
                        pts[:, 0] += slot.ti
                        interp_pts[wf_type] += pts.tolist()

        if time[-1] < total_duration - 1:
            time += [time[-1] + 1, total_duration - 1]
            amp += [0, 0]
            detuning += [0, 0]
        # Store everything
        time.pop(0)  # Removes the -1 in the beginning
        data[ch] = {
            "time": time,
            "amp": amp,
            "detuning": detuning,
            "target": target,
        }
        if hasattr(seq, "_measurement"):
            data[ch]["measurement"] = seq._measurement
        if interp_pts:
            data[ch]["interp_pts"] = interp_pts
    data["total_duration"] = total_duration
    return data


def draw_sequence(
    seq: pulser.sequence.Sequence,
    sampling_rate: Optional[float] = None,
    draw_phase_area: bool = False,
    draw_interp_pts: bool = True,
    draw_phase_shifts: bool = False,
) -> None:
    """Draws the entire sequence.

    Args:
        seq (pulser.Sequence): The input sequence of operations on a device.
        sampling_rate (float): Sampling rate of the effective pulse used by
            the solver. If present, plots the effective pulse alongside the
            input pulse.
        draw_phase_area (bool): Whether phase and area values need to be shown
            as text on the plot, defaults to False.
        draw_interp_pts (bool): When the sequence has pulses with waveforms of
            type InterpolatedWaveform, draws the points of interpolation on
            top of the respective waveforms (defaults to True).
        draw_phase_shifts (bool): Whether phase shift and reference information
            should be added to the plot, defaults to False.
    """

    def phase_str(phi: float) -> str:
        """Formats a phase value for printing."""
        value = (((phi + np.pi) % (2 * np.pi)) - np.pi) / np.pi
        if value == -1:
            return r"$\pi$"
        elif value == 0:
            return "0"  # pragma: no cover - just for safety
        else:
            return fr"{value:.2g}$\pi$"

    n_channels = len(seq._channels)
    if not n_channels:
        raise RuntimeError("Can't draw an empty sequence.")
    data = gather_data(seq)
    total_duration = data["total_duration"]
    time_scale = 1e3 if total_duration > 1e4 else 1

    # Boxes for qubit and phase text
    q_box = dict(boxstyle="round", facecolor="orange")
    ph_box = dict(boxstyle="round", facecolor="ghostwhite")
    area_ph_box = dict(boxstyle="round", facecolor="ghostwhite", alpha=0.7)

    fig = plt.figure(constrained_layout=False, figsize=(20, 4.5 * n_channels))
    gs = fig.add_gridspec(n_channels, 1, hspace=0.075)

    ch_axes = {}
    for i, (ch, gs_) in enumerate(zip(seq._channels, gs)):
        ax = fig.add_subplot(gs_)
        ax.spines["top"].set_color("none")
        ax.spines["bottom"].set_color("none")
        ax.spines["left"].set_color("none")
        ax.spines["right"].set_color("none")
        ax.tick_params(
            labelcolor="w", top=False, bottom=False, left=False, right=False
        )
        ax.set_ylabel(ch, labelpad=40, fontsize=18)
        subgs = gs_.subgridspec(2, 1, hspace=0.0)
        ax1 = fig.add_subplot(subgs[0, :])
        ax2 = fig.add_subplot(subgs[1, :])
        ch_axes[ch] = (ax1, ax2)
        for j, ax in enumerate(ch_axes[ch]):
            ax.axvline(0, linestyle="--", linewidth=0.5, color="grey")
            if j == 0:
                ax.spines["bottom"].set_visible(False)
            else:
                ax.spines["top"].set_visible(False)

            if i < n_channels - 1 or j == 0:
                ax.tick_params(
                    axis="x",
                    which="both",
                    bottom=True,
                    top=False,
                    labelbottom=False,
                    direction="in",
                )
            else:
                unit = "ns" if time_scale == 1 else r"$\mu s$"
                ax.set_xlabel(f"t ({unit})", fontsize=12)

    if sampling_rate:
        indexes = np.linspace(
            0,
            total_duration - 1,
            int(sampling_rate * total_duration),
            dtype=int,
        )
        times = np.arange(total_duration, dtype=np.double) / time_scale
        solver_time = times[indexes]
        delta_t = np.diff(solver_time)[0]
        # Compare pulse with an interpolated pulse with 100 times more samples
        teff = np.arange(0, max(solver_time), delta_t / 100)

    for ch, (a, b) in ch_axes.items():
        basis = seq._channels[ch].basis
        t = np.array(data[ch]["time"]) / time_scale
        ya = data[ch]["amp"]
        yb = data[ch]["detuning"]
        if sampling_rate:
            t2 = 1
            ya2 = []
            yb2 = []
            for t_solv in solver_time:
                # Find the interval [t[t2],t[t2+1]] containing t_solv
                while t_solv > t[t2]:
                    t2 += 1
                ya2.append(ya[t2])
                yb2.append(yb[t2])
            cs_amp = CubicSpline(solver_time, ya2)
            cs_detuning = CubicSpline(solver_time, yb2)
            yaeff = cs_amp(teff)
            ybeff = cs_detuning(teff)

        t_min = -t[-1] * 0.03
        t_max = t[-1] * 1.05
        a.set_xlim(t_min, t_max)
        b.set_xlim(t_min, t_max)

        max_amp = np.max(ya)
        max_amp = 1 if max_amp == 0 else max_amp
        amp_top = max_amp * 1.2
        a.set_ylim(-0.02, amp_top)
        det_max = np.max(yb)
        det_min = np.min(yb)
        det_range = det_max - det_min
        if det_range == 0:
            det_min, det_max, det_range = -1, 1, 2
        det_top = det_max + det_range * 0.15
        det_bottom = det_min - det_range * 0.05
        b.set_ylim(det_bottom, det_top)

        a.plot(t, ya, color="darkgreen", linewidth=0.8)
        b.plot(t, yb, color="indigo", linewidth=0.8)
        if sampling_rate:
            a.plot(teff, yaeff, color="darkgreen", linewidth=0.8)
            b.plot(teff, ybeff, color="indigo", linewidth=0.8, ls="-")
            a.fill_between(teff, 0, yaeff, color="darkgreen", alpha=0.3)
            b.fill_between(teff, 0, ybeff, color="indigo", alpha=0.3)
        else:
            a.fill_between(t, 0, ya, color="darkgreen", alpha=0.3)
            b.fill_between(t, 0, yb, color="indigo", alpha=0.3)
        a.set_ylabel(r"$\Omega$ (rad/µs)", fontsize=14, labelpad=10)
        b.set_ylabel(r"$\delta$ (rad/µs)", fontsize=14)

        if draw_phase_area:
            top = False  # Variable to track position of box, top or center.
            draw_phase = any(
                seq_.type.phase != 0
                for seq_ in seq._schedule[ch]
                if isinstance(seq_.type, Pulse)
            )
            for pulse_num, seq_ in enumerate(seq._schedule[ch]):
                # Select only `Pulse` objects
                if isinstance(seq_.type, Pulse):
                    if sampling_rate:
                        area_val = (
                            np.sum(
                                cs_amp(
                                    np.arange(seq_.ti, seq_.tf) / time_scale
                                )
                            )
                            * 1e-3
                            / np.pi
                        )
                    else:
                        area_val = seq_.type.amplitude.integral / np.pi
                    phase_val = seq_.type.phase
                    x_plot = (seq_.ti + seq_.tf) / 2 / time_scale
                    if (
                        seq._schedule[ch][pulse_num - 1].type == "target"
                        or not top
                    ):
                        y_plot = np.max(seq_.type.amplitude.samples) / 2
                        top = True  # Next box at the top.
                    elif top:
                        y_plot = np.max(seq_.type.amplitude.samples)
                        top = False  # Next box at the center.
                    area_fmt = (
                        r"A: $\pi$"
                        if round(area_val, 2) == 1
                        else fr"A: {area_val:.2g}$\pi$"
                    )
                    if not draw_phase:
                        txt = area_fmt
                    else:
                        phase_fmt = fr"$\phi$: {phase_str(phase_val)}"
                        txt = "\n".join([phase_fmt, area_fmt])
                    a.text(
                        x_plot,
                        y_plot,
                        txt,
                        fontsize=10,
                        ha="center",
                        va="center",
                        bbox=area_ph_box,
                    )

        target_regions = []  # [[start1, [targets1], end1],...]
        for coords in data[ch]["target"]:
            targets = list(data[ch]["target"][coords])
            tgt_strs = [str(q) for q in targets]
            tgt_txt_y = max_amp * 1.1 - 0.25 * (len(targets) - 1)
            tgt_str = "\n".join(tgt_strs)
            if coords == "initial":
                x = t_min + t[-1] * 0.005
                target_regions.append([0, targets])
                if seq._channels[ch].addressing == "Global":
                    a.text(
                        x,
                        amp_top * 0.98,
                        "GLOBAL",
                        fontsize=13,
                        rotation=90,
                        ha="left",
                        va="top",
                        bbox=q_box,
                    )
                else:
                    a.text(
                        x,
                        tgt_txt_y,
                        tgt_str,
                        fontsize=12,
                        ha="left",
                        bbox=q_box,
                    )
                    phase = seq._phase_ref[basis][targets[0]][0]
                    if phase and draw_phase_shifts:
                        msg = r"$\phi=$" + phase_str(phase)
                        a.text(
                            0,
                            max_amp * 1.1,
                            msg,
                            ha="left",
                            fontsize=12,
                            bbox=ph_box,
                        )
            else:
                ti, tf = np.array(coords) / time_scale
                target_regions[-1].append(ti)  # Closing previous regions
                target_regions.append(
                    [tf + 1 / time_scale, targets]
                )  # New one
                phase = seq._phase_ref[basis][targets[0]][tf * time_scale + 1]
                a.axvspan(ti, tf, alpha=0.4, color="grey", hatch="//")
                b.axvspan(ti, tf, alpha=0.4, color="grey", hatch="//")
                a.text(
                    tf + t[-1] * 5e-3,
                    tgt_txt_y,
                    tgt_str,
                    ha="left",
                    fontsize=12,
                    bbox=q_box,
                )
                if phase and draw_phase_shifts:
                    msg = r"$\phi=$" + phase_str(phase)
                    wrd_len = len(max(tgt_strs, key=len))
                    x = tf + t[-1] * 0.01 * (wrd_len + 1)
                    a.text(
                        x,
                        max_amp * 1.1,
                        msg,
                        ha="left",
                        fontsize=12,
                        bbox=ph_box,
                    )
        # Terminate the last open regions
        if target_regions:
            target_regions[-1].append(t[-1])
        for start, targets_, end in (
            target_regions if draw_phase_shifts else []
        ):
            start = cast(float, start)
            targets_ = cast(list, targets_)
            end = cast(float, end)
            # All targets have the same ref, so we pick
            q = targets_[0]
            ref = seq._phase_ref[basis][q]
            if end != total_duration - 1 or "measurement" not in data[ch]:
                end += 1 / time_scale
            for t_, delta in ref.changes(start, end, time_scale=time_scale):
                conf = dict(linestyle="--", linewidth=1.5, color="black")
                a.axvline(t_, **conf)
                b.axvline(t_, **conf)
                msg = "\u27F2 " + phase_str(delta)
                a.text(
                    t_ - t[-1] * 8e-3,
                    max_amp * 1.1,
                    msg,
                    ha="right",
                    fontsize=14,
                    bbox=ph_box,
                )

        if "measurement" in data[ch]:
            msg = f"Basis: {data[ch]['measurement']}"
            b.text(
                t[-1] * 1.025,
                det_top,
                msg,
                ha="center",
                va="center",
                fontsize=14,
                color="white",
                rotation=90,
            )
            a.axvspan(t[-1], t_max, color="midnightblue", alpha=1)
            b.axvspan(t[-1], t_max, color="midnightblue", alpha=1)
            a.axhline(0, xmax=0.95, linestyle="-", linewidth=0.5, color="grey")
            b.axhline(0, xmax=0.95, linestyle=":", linewidth=0.5, color="grey")
        else:
            a.axhline(0, linestyle="-", linewidth=0.5, color="grey")
            b.axhline(0, linestyle=":", linewidth=0.5, color="grey")

        if "interp_pts" in data[ch] and draw_interp_pts:
            all_points = data[ch]["interp_pts"]
            if "amplitude" in all_points:
                pts = np.array(all_points["amplitude"])
                a.scatter(pts[:, 0], pts[:, 1], color="darkgreen")
            if "detuning" in all_points:
                pts = np.array(all_points["detuning"])
                b.scatter(pts[:, 0], pts[:, 1], color="indigo")

    plt.show()



# Copyright 2020 Pulser Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import numpy as np
from scipy.spatial.distance import pdist

from pulser import Register, Pulse
from pulser.channels import Channel
from pulser.json.utils import obj_to_dict


@dataclass(frozen=True, repr=False)
class Device:
    r"""Definition of a neutral-atom device.

    Attributes:
        name: The name of the device.
        dimensions: Whether it supports 2D or 3D arrays.
        max_atom_num: Maximum number of atoms supported in an array.
        max_radial_distance: The furthest away an atom can be from the center
            of the array (in μm).
        min_atom_distance: The closest together two atoms can be (in μm).
        interaction_coeff: :math:`C_6/\hbar` (in :math:`\mu m^6 / \mu s`),
            which sets the van der Waals interaction strength between atoms in
            the Rydberg state.
    """

    name: str
    dimensions: int
    max_atom_num: int
    max_radial_distance: int
    min_atom_distance: int
    _channels: tuple[tuple[str, Channel], ...]
    interaction_coeff: float = 5008713.0

    def __post_init__(self) -> None:
        # Hack to override the docstring of an instance
        self.__dict__["__doc__"] = self._specs(for_docs=True)

    @property
    def channels(self) -> dict[str, Channel]:
        """Dictionary of available channels on this device."""
        return dict(self._channels)

    @property
    def supported_bases(self) -> set[str]:
        """Available electronic transitions for control and measurement."""
        return {ch.basis for ch in self.channels.values()}

    def print_specs(self) -> None:
        """Prints the device specifications."""
        title = f"{self.name} Specifications"
        header = ["-" * len(title), title, "-" * len(title)]
        print("\n".join(header))
        print(self._specs())

    def __repr__(self) -> str:
        return self.name

    def rydberg_blockade_radius(self, rabi_frequency: float) -> float:
        """Calculates the Rydberg blockade radius for a given Rabi frequency.

        Args:
            rabi_frequency(float): The Rabi frequency, in rad/µs.

        Returns:
            float: The rydberg blockade radius, in μm.
        """
        return (self.interaction_coeff / rabi_frequency) ** (1 / 6)

    def rabi_from_blockade(self, blockade_radius: float) -> float:
        """The maximum Rabi frequency value to enforce a given blockade radius.

        Args:
            blockade_radius(float): The Rydberg blockade radius, in µm.

        Returns:
            float: The maximum rabi frequency value, in rad/µs.
        """
        return self.interaction_coeff / blockade_radius ** 6

    def validate_register(self, register: Register) -> None:
        """Checks if 'register' is compatible with this device.

        Args:
            register(pulser.Register): The Register to validate.
        """
        if not isinstance(register, Register):
            raise TypeError("register has to be a pulser.Register instance.")

        atoms = list(register.qubits.values())
        if len(atoms) > self.max_atom_num:
            raise ValueError(
                f"The number of atoms ({len(atoms)})"
                " must be less than or equal to the maximum"
                " number of atoms supported by this device"
                f" ({self.max_atom_num})."
            )

        if register._dim != self.dimensions:
            raise ValueError(
                f"All qubit positions must be {self.dimensions}D " "vectors."
            )

        if len(atoms) > 1:
            distances = pdist(atoms)  # Pairwise distance between atoms
            if np.min(distances) < self.min_atom_distance:
                raise ValueError(
                    "Qubit positions don't respect the minimal "
                    "distance between atoms for this device."
                )

        if np.max(np.linalg.norm(atoms, axis=1)) > self.max_radial_distance:
            raise ValueError(
                "All qubits must be at most "
                f"{self.max_radial_distance} μm away from the "
                "center of the array."
            )

    def validate_pulse(self, pulse: Pulse, channel_id: str) -> None:
        """Checks if a pulse can be executed on a specific device channel.

        Args:
            pulse (Pulse): The pulse to validate.
            channel_id (str): The channel ID used to index the chosen channel
                on this device.
        """
        ch = self.channels[channel_id]
        if np.any(pulse.amplitude.samples > ch.max_amp):
            raise ValueError(
                "The pulse's amplitude goes over the maximum "
                "value allowed for the chosen channel."
            )
        if np.any(
            np.round(np.abs(pulse.detuning.samples), decimals=6)
            > ch.max_abs_detuning
        ):
            raise ValueError(
                "The pulse's detuning values go out of the range "
                "allowed for the chosen channel."
            )

    def _specs(self, for_docs: bool = False) -> str:
        lines = [
            "\nRegister requirements:",
            f" - Dimensions: {self.dimensions}D",
            f" - Maximum number of atoms: {self.max_atom_num}",
            f" - Maximum distance from origin: {self.max_radial_distance} μm",
            (
                " - Minimum distance between neighbouring atoms: "
                f"{self.min_atom_distance} μm"
            ),
            (
                r" - Interaction coefficient (:math:`C_6/\hbar`): "
                fr"{self.interaction_coeff} :math:`\mu m^6 / \mu s`"
            ),
            "\nChannels:",
        ]

        ch_lines = []
        for name, ch in self._channels:
            if for_docs:
                ch_lines += [
                    f" - ID: '{name}'",
                    f"\t- Type: {ch.name} (*{ch.basis}* basis)",
                    f"\t- Addressing: {ch.addressing}",
                    (
                        "\t"
                        + r"- Maximum :math:`\Omega`:"
                        + f" {ch.max_amp:.4g} rad/µs"
                    ),
                    (
                        "\t"
                        + r"- Maximum :math:`|\delta|`:"
                        + f" {ch.max_abs_detuning:.4g} rad/µs"
                    ),
                ]
                if ch.addressing == "Local":
                    ch_lines += [
                        f"\t- Maximum time to retarget: {ch.retarget_time} ns",
                        f"\t- Maximum simultaneous targets: {ch.max_targets}",
                    ]
                ch_lines += [
                    f"\t- Clock period: {ch.clock_period} ns",
                    f"\t- Minimum instruction duration: {ch.min_duration} ns",
                ]
            else:
                ch_lines.append(f" - '{name}': {ch!r}")

        return "\n".join(lines + ch_lines)

    def _to_dict(self) -> dict[str, Any]:
        return obj_to_dict(
            self, _build=False, _module="pulser.devices", _name=self.name
        )
